import dataclasses
from functools import cached_property
from typing import Any, Dict, Optional, Union

from eth_abi import encode as abi_encode
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes

from gnosis.eth.utils import fast_keccak, fast_to_checksum_address


@dataclasses.dataclass(eq=True, frozen=True)
class UserOperationMetadata:
    transaction_hash: bytes
    block_hash: bytes
    block_number: int


@dataclasses.dataclass(eq=True, frozen=True)
class UserOperation:
    """
    EIP4337 UserOperation for Entrypoint v0.6

    https://github.com/eth-infinitism/account-abstraction/blob/v0.6.0/contracts/interfaces/UserOperation.sol
    """

    user_operation_hash: bytes
    sender: ChecksumAddress
    nonce: int
    init_code: bytes
    call_data: bytes
    call_gas_limit: int
    verification_gas_limit: int
    pre_verification_gas: int
    max_fee_per_gas: int
    max_priority_fee_per_gas: int
    paymaster_and_data: bytes
    signature: bytes
    entry_point: ChecksumAddress
    metadata: Optional[UserOperationMetadata] = None

    @classmethod
    def from_bundler_response(
        cls,
        user_operation_hash: Union[HexStr, bytes],
        user_operation_response: Dict[str, Any],
    ) -> Union["UserOperation", "UserOperationV07"]:
        user_operation = user_operation_response["userOperation"]
        metadata = UserOperationMetadata(
            HexBytes(user_operation_response["transactionHash"]),
            HexBytes(user_operation_response["blockHash"]),
            int(user_operation_response["blockNumber"], 16),
        )
        if "initCode" in user_operation:
            return cls(
                HexBytes(user_operation_hash),
                ChecksumAddress(user_operation["sender"]),
                int(user_operation["nonce"], 16),
                HexBytes(user_operation["initCode"]),
                HexBytes(user_operation["callData"]),
                int(user_operation["callGasLimit"], 16),
                int(user_operation["verificationGasLimit"], 16),
                int(user_operation["preVerificationGas"], 16),
                int(user_operation["maxFeePerGas"], 16),
                int(user_operation["maxPriorityFeePerGas"], 16),
                HexBytes(user_operation["paymasterAndData"]),
                HexBytes(user_operation["signature"]),
                ChecksumAddress(user_operation_response["entryPoint"]),
                metadata=metadata,
            )
        else:
            if paymaster := user_operation.get("paymaster"):
                # Paymaster parameters are optional
                paymaster_verification_gas_limit = int(
                    user_operation["paymasterVerificationGasLimit"], 16
                )
                paymaster_post_op_gas_limit = int(
                    user_operation["paymasterPostOpGasLimit"], 16
                )
                paymaster_data = HexBytes(user_operation["paymasterData"])
            else:
                paymaster_verification_gas_limit = None
                paymaster_post_op_gas_limit = None
                paymaster_data = None

            return UserOperationV07(
                HexBytes(user_operation_hash),
                ChecksumAddress(user_operation["sender"]),
                int(user_operation["nonce"], 16),
                ChecksumAddress(user_operation["factory"]),
                HexBytes(user_operation["factoryData"]),
                HexBytes(user_operation["callData"]),
                int(user_operation["callGasLimit"], 16),
                int(user_operation["verificationGasLimit"], 16),
                int(user_operation["preVerificationGas"], 16),
                int(user_operation["maxPriorityFeePerGas"], 16),
                int(user_operation["maxFeePerGas"], 16),
                HexBytes(user_operation["signature"]),
                ChecksumAddress(user_operation_response["entryPoint"]),
                paymaster_verification_gas_limit,
                paymaster_post_op_gas_limit,
                paymaster,
                paymaster_data,
                metadata=metadata,
            )

    def __str__(self):
        return f"User Operation sender={self.sender} nonce={self.nonce} hash={self.user_operation_hash.hex()}"

    @cached_property
    def paymaster(self) -> Optional[ChecksumAddress]:
        if self.paymaster_and_data and len(self.paymaster_and_data) >= 20:
            return fast_to_checksum_address(self.paymaster_and_data[:20])
        return None

    @cached_property
    def paymaster_data(self) -> Optional[bytes]:
        if self.paymaster_and_data:
            return self.paymaster_and_data[20:]
        return None

    def calculate_user_operation_hash(self, chain_id: int) -> bytes:
        hash_init_code = fast_keccak(self.init_code)
        hash_call_data = fast_keccak(self.call_data)
        hash_paymaster_and_data = fast_keccak(self.paymaster_and_data)
        user_operation_encoded = abi_encode(
            [
                "address",
                "uint256",
                "bytes32",
                "bytes32",
                "uint256",
                "uint256",
                "uint256",
                "uint256",
                "uint256",
                "bytes32",
            ],
            [
                self.sender,
                self.nonce,
                hash_init_code,
                hash_call_data,
                self.call_gas_limit,
                self.verification_gas_limit,
                self.pre_verification_gas,
                self.max_fee_per_gas,
                self.max_priority_fee_per_gas,
                hash_paymaster_and_data,
            ],
        )
        return fast_keccak(
            abi_encode(
                ["bytes32", "address", "uint256"],
                [fast_keccak(user_operation_encoded), self.entry_point, chain_id],
            )
        )


@dataclasses.dataclass(eq=True, frozen=True)
class UserOperationV07:
    """
    EIP4337 UserOperation for Entrypoint v0.7

    https://github.com/eth-infinitism/account-abstraction/blob/v0.7.0/contracts/interfaces/PackedUserOperation.sol
    """

    user_operation_hash: bytes
    sender: ChecksumAddress
    nonce: int
    factory: ChecksumAddress
    factory_data: bytes
    call_data: bytes
    call_gas_limit: int
    verification_gas_limit: int
    pre_verification_gas: int
    max_priority_fee_per_gas: int
    max_fee_per_gas: int
    signature: bytes
    entry_point: ChecksumAddress
    paymaster_verification_gas_limit: Optional[int] = None
    paymaster_post_op_gas_limit: Optional[int] = None
    paymaster: Optional[bytes] = None
    paymaster_data: Optional[bytes] = None
    metadata: Optional[UserOperationMetadata] = None

    @property
    def account_gas_limits(self) -> bytes:
        """
        :return:Account Gas Limits is a `bytes32` in Solidity, first `bytes16` `verification_gas_limit` and then `call_gas_limit`
        """
        return HexBytes(self.verification_gas_limit).rjust(16, b"\x00") + HexBytes(
            self.call_gas_limit
        ).rjust(16, b"\x00")

    @property
    def gas_fees(self) -> bytes:
        """
        :return: Gas Fees is a `bytes32` in Solidity, first `bytes16` `verification_gas_limit` and then `call_gas_limit`
        """
        return HexBytes(self.max_priority_fee_per_gas).rjust(16, b"\x00") + HexBytes(
            self.max_fee_per_gas
        ).rjust(16, b"\x00")

    @property
    def paymaster_and_data(self) -> bytes:
        if not self.paymaster:
            return b""
        return (
            HexBytes(self.paymaster).rjust(20, b"\x00")
            + HexBytes(self.paymaster_verification_gas_limit).rjust(16, b"\x00")
            + HexBytes(self.paymaster_post_op_gas_limit).rjust(16, b"\x00")
            + HexBytes(self.paymaster_data)
        )

    def calculate_user_operation_hash(self, chain_id: int) -> bytes:
        hash_init_code = fast_keccak(HexBytes(self.factory) + self.factory_data)
        hash_call_data = fast_keccak(self.call_data)
        hash_paymaster_and_data = fast_keccak(self.paymaster_and_data)
        user_operation_encoded = abi_encode(
            [
                "address",
                "uint256",
                "bytes32",
                "bytes32",
                "bytes32",
                "uint256",
                "bytes32",
                "bytes32",
            ],
            [
                self.sender,
                self.nonce,
                hash_init_code,
                hash_call_data,
                self.account_gas_limits,
                self.pre_verification_gas,
                self.gas_fees,
                hash_paymaster_and_data,
            ],
        )
        return fast_keccak(
            abi_encode(
                ["bytes32", "address", "uint256"],
                [fast_keccak(user_operation_encoded), self.entry_point, chain_id],
            )
        )



import os
from dataclasses import dataclass
from functools import cache
from typing import Any, Dict, List, Optional, Union

import requests
from eth_typing import HexStr
from hexbytes import HexBytes


class EnsClient:
    """
    Resolves Ethereum Name Service domains using ``thegraph`` API
    """

    @dataclass
    class Config:
        base_url: str

        @property
        def url(self) -> str:
            return self.base_url

    @dataclass
    class SubgraphConfig(Config):
        api_key: str
        subgraph_id: str

        @property
        def url(self):
            return f"{self.base_url}/api/{self.api_key}/subgraphs/id/{self.subgraph_id}"

    def __init__(self, config: Config):
        self.config = config
        self.request_timeout = int(
            os.environ.get("ENS_CLIENT_REQUEST_TIMEOUT", 5)
        )  # Seconds
        self.request_session = requests.Session()

    def is_available(self) -> bool:
        """
        :return: True if service is available, False if it's down
        """
        try:
            return self.request_session.get(
                self.config.url, timeout=self.request_timeout
            ).ok
        except IOError:
            return False

    @staticmethod
    def domain_hash_to_hex_str(domain_hash: Union[HexStr, bytes, int]) -> HexStr:
        """
        :param domain_hash:
        :return: Domain hash as an hex string of 66 chars (counting with 0x), padding with zeros if needed
        """
        if not domain_hash:
            domain_hash = b""
        return HexStr("0x" + HexBytes(domain_hash).hex()[2:].rjust(64, "0"))

    @cache
    def _query_by_domain_hash(self, domain_hash_str: HexStr) -> Optional[str]:
        query = """
                {
                    domains(where: {labelhash: "domain_hash"}) {
                        labelName
                    }
                }
                """.replace(
            "domain_hash", domain_hash_str
        )
        try:
            response = self.request_session.post(
                self.config.url,
                json={"query": query},
                timeout=self.request_timeout,
            )
        except IOError:
            return None

        """
        Example:
        {
            "data": {
                "domains": [
                    {
                        "labelName": "safe-multisig"
                    }
                ]
            }
        }
        """
        if response.ok:
            data = response.json()
            if data:
                domains = data.get("data", {}).get("domains")
                if domains:
                    return domains[0].get("labelName")
        return None

    def query_by_domain_hash(
        self, domain_hash: Union[HexStr, bytes, int]
    ) -> Optional[str]:
        """
        Get domain label from domain_hash (keccak of domain name without the TLD, don't confuse with namehash)
        used for ENS ERC721 token_id. Use another method for caching purposes (use same parameter type)

        :param domain_hash: keccak of domain name without the TLD, don't confuse with namehash. E.g. For
            batman.eth it would be just keccak('batman')
        :return: domain label if found
        """
        domain_hash_str = self.domain_hash_to_hex_str(domain_hash)
        return self._query_by_domain_hash(domain_hash_str)

    def query_by_account(self, account: str) -> Optional[List[Dict[str, Any]]]:
        """
        :param account: ethereum account to search for ENS registered addresses
        :return: None if there's a problem or not found, otherwise example of dictionary returned:
        {
            "registrations": [
                {
                    "domain": {
                        "isMigrated": true,
                        "labelName": "gilfoyle",
                        "labelhash": "0xadfd886b420023026d5c0b1be0ffb5f18bb2f37143dff545aeaea0d23a4ba910",
                        "name": "gilfoyle.eth",
                        "parent": {
                            "name": "eth"
                        }
                    },
                    "expiryDate": "1905460880"
                }
            ]
        }
        """
        query = """query getRegistrations {
          account(id: "account_id") {
            registrations {
              expiryDate
              domain {
                labelName
                labelhash
                name
                isMigrated
                parent {
                  name
                }
              }
            }
          }
        }""".replace(
            "account_id", account.lower()
        )
        try:
            response = self.request_session.post(
                self.config.url,
                json={"query": query},
                timeout=self.request_timeout,
            )
        except IOError:
            return None

        if response.ok:
            data = response.json()
            if data:
                return data.get("data", {}).get("account")
        return None



import functools
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import List, Optional, Tuple

import requests
from eth_abi import decode as decode_abi
from eth_abi.exceptions import DecodingError
from eth_abi.packed import encode_packed
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.contract import Contract
from web3.exceptions import Web3Exception

from .. import EthereumClient, EthereumNetwork
from ..constants import NULL_ADDRESS
from ..contracts import (
    get_erc20_contract,
    get_uniswap_factory_contract,
    get_uniswap_v2_factory_contract,
    get_uniswap_v2_pair_contract,
    get_uniswap_v2_router_contract,
)
from ..utils import fast_bytes_to_checksum_address, fast_keccak, get_empty_tx_params
from .abis.aave_abis import AAVE_ATOKEN_ABI
from .abis.balancer_abis import balancer_pool_abi
from .abis.cream_abis import cream_ctoken_abi
from .abis.mooniswap_abis import mooniswap_abi
from .abis.zerion_abis import ZERION_TOKEN_ADAPTER_ABI
from .exceptions import CannotGetPriceFromOracle, InvalidPriceFromOracle
from .helpers.curve_gauge_list import CURVE_GAUGE_TO_LP_TOKEN
from .utils import get_decimals

logger = logging.getLogger(__name__)


@dataclass
class UnderlyingToken:
    address: ChecksumAddress
    quantity: int


class BaseOracle(ABC):
    @classmethod
    @abstractmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        raise NotImplementedError


class PriceOracle(BaseOracle):
    @abstractmethod
    def get_price(self, *args) -> float:
        raise NotImplementedError


class PricePoolOracle(BaseOracle):
    @abstractmethod
    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        raise NotImplementedError


class ComposedPriceOracle(BaseOracle):
    @abstractmethod
    def get_underlying_tokens(self, *args) -> List[Tuple[UnderlyingToken]]:
        raise NotImplementedError


class UniswapOracle(PriceOracle):
    """
    Uniswap V1 Oracle

    https://docs.uniswap.org/protocol/V1/guides/connect-to-uniswap
    """

    ADDRESSES = {
        EthereumNetwork.MAINNET: "0xc0a47dFe034B400B47bDaD5FecDa2621de6c4d95",
        EthereumNetwork.RINKEBY: "0xf5D915570BC477f9B8D6C0E980aA81757A3AaC36",
        EthereumNetwork.ROPSTEN: "0x9c83dCE8CA20E9aAF9D3efc003b2ea62aBC08351",
        EthereumNetwork.GOERLI: "0x6Ce570d02D73d4c384b46135E87f8C592A8c86dA",
    }

    def __init__(
        self,
        ethereum_client: EthereumClient,
        uniswap_factory_address: Optional[str] = None,
    ):
        """
        :param ethereum_client:
        :param uniswap_factory_address: https://docs.uniswap.io/frontend-integration/connect-to-uniswap#factory-contract
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self._uniswap_factory_address = uniswap_factory_address

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() in cls.ADDRESSES

    @cached_property
    def uniswap_factory_address(self):
        if self._uniswap_factory_address:
            return self._uniswap_factory_address
        return self.ADDRESSES.get(
            self.ethereum_client.get_network(),
            self.ADDRESSES.get(EthereumNetwork.MAINNET),
        )

    @cached_property
    def uniswap_factory(self):
        return get_uniswap_factory_contract(self.w3, self.uniswap_factory_address)

    @functools.lru_cache(maxsize=2048)
    def get_uniswap_exchange(self, token_address: str) -> str:
        return self.uniswap_factory.functions.getExchange(token_address).call()

    def _get_balances_without_batching(
        self, uniswap_exchange_address: str, token_address: str
    ):
        balance = self.ethereum_client.get_balance(uniswap_exchange_address)
        token_decimals = self.ethereum_client.erc20.get_decimals(token_address)
        token_balance = self.ethereum_client.erc20.get_balance(
            uniswap_exchange_address, token_address
        )
        return balance, token_decimals, token_balance

    def _get_balances_using_batching(
        self, uniswap_exchange_address: str, token_address: str
    ):
        # Use batching instead
        payload_balance = {
            "jsonrpc": "2.0",
            "method": "eth_getBalance",
            "params": [uniswap_exchange_address, "latest"],
            "id": 0,
        }
        erc20 = get_erc20_contract(self.w3, token_address)
        params = get_empty_tx_params()
        decimals_data = erc20.functions.decimals().build_transaction(params)["data"]
        token_balance_data = erc20.functions.balanceOf(
            uniswap_exchange_address
        ).build_transaction(params)["data"]
        datas = [decimals_data, token_balance_data]
        payload_calls = [
            {
                "id": i + 1,
                "jsonrpc": "2.0",
                "method": "eth_call",
                "params": [{"to": token_address, "data": data}, "latest"],
            }
            for i, data in enumerate(datas)
        ]
        payloads = [payload_balance] + payload_calls
        r = requests.post(
            self.ethereum_client.ethereum_node_url,
            json=payloads,
            timeout=self.ethereum_client.timeout,
        )
        if not r.ok:
            raise CannotGetPriceFromOracle(
                f"Error from node with url={self.ethereum_client.ethereum_node_url}"
            )

        results = []
        for result in r.json():
            if "result" not in result:
                raise CannotGetPriceFromOracle(result["error"])
            else:
                results.append(HexBytes(result["result"]))

        balance = int(results[0].hex(), 16)
        token_decimals = decode_abi(["uint8"], results[1])[0]
        token_balance = decode_abi(["uint256"], results[2])[0]
        return balance, token_decimals, token_balance

    def get_price(self, token_address: str) -> float:
        try:
            uniswap_exchange_address = self.get_uniswap_exchange(token_address)
            if uniswap_exchange_address == NULL_ADDRESS:
                raise ValueError
        except (Web3Exception, DecodingError, ValueError) as e:
            message = f"Non existing uniswap exchange for token={token_address}"
            logger.debug(message)
            raise CannotGetPriceFromOracle(message) from e

        try:
            balance, token_decimals, token_balance = self._get_balances_using_batching(
                uniswap_exchange_address, token_address
            )
            # Check liquidity. Require at least 2 ether to be on the pool
            if balance / 1e18 < 2:
                raise CannotGetPriceFromOracle(
                    f"Not enough liquidity for token={token_address}"
                )

            price = balance / token_balance / 10 ** (18 - token_decimals)
            if price <= 0.0:
                message = (
                    f"price={price} <= 0 from uniswap-factory={uniswap_exchange_address} "
                    f"for token={token_address}"
                )
                logger.debug(message)
                raise InvalidPriceFromOracle(message)
            return price
        except (
            Web3Exception,
            DecodingError,
            ValueError,
            ZeroDivisionError,
        ) as e:
            message = f"Cannot get token balance for token={token_address}"
            logger.debug(message)
            raise CannotGetPriceFromOracle(message) from e


class UniswapV2Oracle(PricePoolOracle, PriceOracle):
    ROUTER_ADDRESSES = {
        EthereumNetwork.MAINNET: "0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D",
    }

    # Pair init code is keccak(getCode(UniswapV2Pair))
    PAIR_INIT_CODE = HexBytes(
        "0x96e8ac4277198ff8b6f785478aa9a39f403cb768dd02cbee326c3e7da348845f"
    )

    def __init__(
        self, ethereum_client: EthereumClient, router_address: Optional[str] = None
    ):
        """
        :param ethereum_client:
        :param router_address: https://uniswap.org/docs/v2/smart-contracts/router02/
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.router_address: str = router_address or self.ROUTER_ADDRESSES.get(
            ethereum_client.get_network(),
            self.ROUTER_ADDRESSES[EthereumNetwork.MAINNET],
        )
        self.router = get_uniswap_v2_router_contract(
            ethereum_client.w3, self.router_address
        )

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.is_contract(
            cls.ROUTER_ADDRESSES.get(
                ethereum_client.get_network(),
                cls.ROUTER_ADDRESSES[EthereumNetwork.MAINNET],
            )
        )

    @cached_property
    def factory(self):
        return get_uniswap_v2_factory_contract(
            self.ethereum_client.w3, self.factory_address
        )

    @cached_property
    def factory_address(self) -> str:
        """
        :return: Uniswap factory checksummed address
        :raises: Web3Exception: If router contract is not deployed
        """
        return self.router.functions.factory().call()

    @cached_property
    def weth_address(self) -> str:
        """
        :return: Wrapped ether checksummed address
        :raises: Web3Exception: If router contract is not deployed
        """
        return self.router.functions.WETH().call()

    @functools.lru_cache(maxsize=2048)
    def get_pair_address(
        self, token_address: str, token_address_2: str
    ) -> Optional[str]:
        """
        Get uniswap pair address. `token_address` and `token_address_2` are interchangeable.
        https://uniswap.org/docs/v2/smart-contracts/factory/

        :param token_address:
        :param token_address_2:
        :return: Address of the pair for `token_address` and `token_address_2`, if it has been created, else `None`.
        """
        # Token order does not matter for getting pair, just for creating or querying PairCreated event
        pair_address = self.factory.functions.getPair(
            token_address, token_address_2
        ).call()
        if pair_address == NULL_ADDRESS:
            return None
        return pair_address

    @functools.lru_cache(maxsize=2048)
    def calculate_pair_address(self, token_address: str, token_address_2: str):
        """
        Calculate pair address without querying blockchain.
        https://uniswap.org/docs/v2/smart-contract-integration/getting-pair-addresses/#docs-header


        :param token_address:
        :param token_address_2:
        :return: Checksummed address for token pair. It could be not created yet
        """
        if token_address.lower() > token_address_2.lower():
            token_address, token_address_2 = token_address_2, token_address
        salt = fast_keccak(
            encode_packed(["address", "address"], [token_address, token_address_2])
        )
        address = fast_keccak(
            encode_packed(
                ["bytes", "address", "bytes", "bytes"],
                [HexBytes("ff"), self.factory_address, salt, self.PAIR_INIT_CODE],
            )
        )[-20:]
        return fast_bytes_to_checksum_address(address)

    def get_reserves(self, pair_address: str) -> Tuple[int, int]:
        """
        Returns the number of tokens in the pool. `getReserves()` also returns the block.timestamp (mod 2**32) of
        the last block during which an interaction occurred for the pair, but it's ignored.
        https://uniswap.org/docs/v2/smart-contracts/pair/

        :return: Reserves of `token_address` and `token_address_2` used to price trades and distribute liquidity.
        """
        pair_contract = get_uniswap_v2_pair_contract(
            self.ethereum_client.w3, pair_address
        )
        # Reserves return token_1 reserves, token_2 reserves and block timestamp (mod 2**32) of last interaction
        reserves_1, reserves_2, _ = pair_contract.functions.getReserves().call()
        return reserves_1, reserves_2

    def get_price(
        self, token_address: str, token_address_2: Optional[str] = None
    ) -> float:
        # These lines only make sense when `get_pair_address` is used. `calculate_pair_address` will always return
        # an address, even it that exchange is not deployed

        # pair_address = self.get_pair_address(token_address, token_address_2)
        # if not pair_address:
        #    error_message = f'Non existing uniswap V2 exchange for token={token_address}'
        #    logger.warning(error_message)
        #    raise CannotGetPriceFromOracle(error_message)
        try:
            token_address_2 = token_address_2 if token_address_2 else self.weth_address
            if token_address == token_address_2:
                return 1.0
            pair_address = self.calculate_pair_address(token_address, token_address_2)
            # Tokens are sorted, so token_1 < token_2
            reserves_1, reserves_2 = self.get_reserves(pair_address)
            decimals_1 = get_decimals(token_address, self.ethereum_client)
            decimals_2 = get_decimals(token_address_2, self.ethereum_client)
            if token_address.lower() > token_address_2.lower():
                reserves_2, reserves_1 = reserves_1, reserves_2

            # Check liquidity. Require at least 2 units of every token to be on the pool
            if reserves_1 / 10**decimals_1 < 2 or reserves_2 / 10**decimals_2 < 2:
                raise CannotGetPriceFromOracle(
                    f"Not enough liquidity for pair token_1={token_address} "
                    f"token_2={token_address_2}"
                )
            decimals_normalized_reserves_1 = reserves_1 * 10**decimals_2
            decimals_normalized_reserves_2 = reserves_2 * 10**decimals_1

            return decimals_normalized_reserves_2 / decimals_normalized_reserves_1
        except (
            Web3Exception,
            DecodingError,
            ValueError,
            ZeroDivisionError,
        ) as e:
            message = (
                f"Cannot get uniswap v2 price for pair token_1={token_address} "
                f"token_2={token_address_2}"
            )
            logger.debug(message)
            raise CannotGetPriceFromOracle(message) from e

    def get_price_without_exception(
        self, token_address: str, token_address_2: Optional[str] = None
    ) -> float:
        """
        :param token_address:
        :param token_address_2:
        :return: Call `get_price`, return 0. instead on an exception if there's any issue
        """
        try:
            return self.get_price(token_address, token_address_2=token_address_2)
        except CannotGetPriceFromOracle:
            return 0.0

    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate pool token price based on its components

        :param pool_token_address:
        :return: Pool token eth price per unit (total pool token supply / 1e18)
        :raises: CannotGetPriceFromOracle
        """
        try:
            pair_contract = get_uniswap_v2_pair_contract(
                self.ethereum_client.w3, pool_token_address
            )
            (
                (reserves_1, reserves_2, _),
                token_address_1,
                token_address_2,
                total_supply,
            ) = self.ethereum_client.batch_call(
                [
                    pair_contract.functions.getReserves(),
                    pair_contract.functions.token0(),
                    pair_contract.functions.token1(),
                    pair_contract.functions.totalSupply(),
                ]
            )
            decimals_1 = get_decimals(token_address_1, self.ethereum_client)
            decimals_2 = get_decimals(token_address_2, self.ethereum_client)

            # Total value for one token should be the same than total value for the other token
            # if pool is under active arbitrage. We use the price for the first token we find
            for token_address, decimals, reserves in zip(
                (token_address_1, token_address_2),
                (decimals_1, decimals_2),
                (reserves_1, reserves_2),
            ):
                try:
                    price = self.get_price(token_address)
                    total_value = (reserves / 10**decimals_1) * price
                    return (total_value * 2) / (total_supply / 1e18)
                except CannotGetPriceFromOracle:
                    continue
        except (
            Web3Exception,
            DecodingError,
            ValueError,
            ZeroDivisionError,
        ) as e:
            message = f"Cannot get uniswap v2 price for pool token={pool_token_address}"
            logger.debug(message)
            raise CannotGetPriceFromOracle(message) from e


class AaveOracle(PriceOracle):
    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of Aave Tokens, UniswapV2 is
        recommended
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() == EthereumNetwork.MAINNET

    def get_price(self, token_address: str) -> float:
        if (
            token_address == "0x4da27a545c0c5B758a6BA100e3a049001de870f5"
        ):  # Stacked Aave
            return self.price_oracle.get_price(
                "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9"
            )
        try:
            underlying_token = (
                self.w3.eth.contract(token_address, abi=AAVE_ATOKEN_ABI)
                .functions.UNDERLYING_ASSET_ADDRESS()
                .call()
            )
            return self.price_oracle.get_price(underlying_token)
        except (Web3Exception, DecodingError, ValueError):
            raise CannotGetPriceFromOracle(
                f"Cannot get price for {token_address}. It is not an Aaave atoken"
            )


class CreamOracle(PriceOracle):
    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of Cream Tokens, UniswapV2 is
        recommended
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() == EthereumNetwork.MAINNET

    def get_price(self, token_address: str) -> float:
        try:
            underlying_token = (
                self.w3.eth.contract(token_address, abi=cream_ctoken_abi)
                .functions.underlying()
                .call()
            )
            return self.price_oracle.get_price(underlying_token)
        except (Web3Exception, DecodingError, ValueError):
            raise CannotGetPriceFromOracle(
                f"Cannot get price for {token_address}. It is not a Cream cToken"
            )


class ZerionComposedOracle(ComposedPriceOracle):
    ZERION_ADAPTER_ADDRESS = None

    def __init__(
        self,
        ethereum_client: EthereumClient,
        zerion_adapter_address: Optional[str] = None,
    ):
        """
        :param ethereum_client:
        :param zerion_adapter_address: Can be retrieved using the Zerion registry on
            0x06FE76B2f432fdfEcAEf1a7d4f6C3d41B5861672 . https://github.com/zeriontech/defi-sdk/wiki/Addresses is
            outdated
        """

        self.zerion_adapter_address = (
            zerion_adapter_address or self.ZERION_ADAPTER_ADDRESS
        )
        if not self.zerion_adapter_address:
            raise ValueError("Expected a Zerion adapter address")
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() == EthereumNetwork.MAINNET

    @cached_property
    def zerion_adapter_contract(self) -> Optional[Contract]:
        """
        :return: https://curve.readthedocs.io/registry-registry.html
        """
        if self.ethereum_client.is_contract(self.zerion_adapter_address):
            return self.w3.eth.contract(
                self.zerion_adapter_address, abi=ZERION_TOKEN_ADAPTER_ABI
            )

    def get_underlying_tokens(
        self, token_address: ChecksumAddress
    ) -> List[UnderlyingToken]:
        """
        Use Zerion Token adapter to return underlying components for pool

        :param token_address: Pool token address
        :return: Price per share and underlying token
        :raises: CannotGetPriceFromOracle
        """
        if not self.zerion_adapter_contract:
            raise CannotGetPriceFromOracle(
                f"{self.__class__.__name__}: Cannot get price for {token_address}. Cannot find Zerion adapter"
            )

        try:
            results = self.zerion_adapter_contract.functions.getComponents(
                token_address
            ).call()
            if results:
                underlying_tokens = []
                for token_address, _, quantity in results:
                    # If there's just one component, quantity must be 100%
                    normalized_quantity = (
                        1.0
                        if len(results) == 1
                        else quantity / 10 ** len(str(quantity))
                    )
                    underlying_tokens.append(
                        UnderlyingToken(token_address, normalized_quantity)
                    )
                return underlying_tokens
        except (ValueError, Web3Exception, DecodingError):
            pass

        raise CannotGetPriceFromOracle(
            f"{self.__class__.__name__}: Cannot get price for {token_address}. It is not a Zerion supported pool token"
        )


class CurveOracle(ZerionComposedOracle):
    """
    Curve pool Oracle. More info on https://curve.fi/
    """

    ZERION_ADAPTER_ADDRESS = (
        "0x99b0bEadc3984eab9842AF81f9fad0C2219108cc"  # Mainnet address
    )

    def get_underlying_tokens(
        self, token_address: ChecksumAddress
    ) -> List[UnderlyingToken]:
        """
        Check if passed token address is a Curve gauge deposit token, if it's a gauge we replace the address with
        the corresponding LP token address
        More info on https://resources.curve.fi/base-features/understanding-gauges
        """
        if CURVE_GAUGE_TO_LP_TOKEN.get(token_address):
            return super().get_underlying_tokens(CURVE_GAUGE_TO_LP_TOKEN[token_address])

        return super().get_underlying_tokens(token_address)


class PoolTogetherOracle(ZerionComposedOracle):
    """
    PoolTogether pool Oracle. More info on https://pooltogether.com/
    """

    ZERION_ADAPTER_ADDRESS = (
        "0xb4E0E1672fFd9b128784dB9f3BE9158fac3f1DFc"  # Mainnet address
    )


class EnzymeOracle(ZerionComposedOracle):
    """
    Enzyme pool Oracle. More info on https://enzyme.finance/
    """

    ZERION_ADAPTER_ADDRESS = (
        "0x9e71455D748C23566b19493D09435574097C7D67"  # Mainnet address
    )


class YearnOracle(ComposedPriceOracle):
    """
    Yearn oracle. More info on https://docs.yearn.finance
    """

    def __init__(
        self,
        ethereum_client: EthereumClient,
        yearn_vault_token_adapter: Optional[
            str
        ] = "0xb460FcC1B6c1CBD7D03F47B6BD5F03994d286c75",
        iearn_token_adapter: Optional[
            str
        ] = "0x65B23774daE2a5be02dD275918DDF048d177a5B4",
    ):
        """
        :param ethereum_client:
        :param yearn_vault_token_adapter:
        :param iearn_token_adapter:
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.yearn_vault_token_adapter = ZerionComposedOracle(
            ethereum_client, yearn_vault_token_adapter
        )
        self.iearn_token_adapter = ZerionComposedOracle(
            ethereum_client, iearn_token_adapter
        )

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() == EthereumNetwork.MAINNET

    def get_underlying_tokens(
        self, token_address: ChecksumAddress
    ) -> List[Tuple[float, ChecksumAddress]]:
        """
        :param token_address:
        :return: Price per share and underlying token
        :raises: CannotGetPriceFromOracle
        """
        for adapter in (self.yearn_vault_token_adapter, self.iearn_token_adapter):
            try:
                # Getting underlying token function is the same for both yVault and yToken
                return adapter.get_underlying_tokens(token_address)
            except CannotGetPriceFromOracle:
                pass

        raise CannotGetPriceFromOracle(
            f"Cannot get price for {token_address}. It is not a Yearn yToken/yVault"
        )


class BalancerOracle(PricePoolOracle):
    """
    Oracle for Balancer. More info on https://balancer.exchange
    """

    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of the Balancer Pool, UniswapV2 is
        recommended
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() == EthereumNetwork.MAINNET

    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate balancer pool token price based on its components

        :param pool_token_address: Balancer pool token address
        :return: Eth price for pool token
        :raises: CannotGetPriceFromOracle
        """
        try:
            balancer_pool_contract = self.w3.eth.contract(
                pool_token_address, abi=balancer_pool_abi
            )
            current_tokens, total_supply = self.ethereum_client.batch_call(
                [
                    balancer_pool_contract.functions.getCurrentTokens(),
                    balancer_pool_contract.functions.totalSupply(),
                ]
            )
            if not current_tokens:
                raise ValueError
            number_tokens = len(current_tokens)
            # denormalized_weight = self.ethereum_client.batch_call([
            #    balancer_pool_contract.functions.getReserves(current_token)
            #    for current_token in current_tokens
            # ])
            token_balances_and_decimals = self.ethereum_client.batch_call(
                [
                    balancer_pool_contract.functions.getBalance(token_address)
                    for token_address in current_tokens
                ]
                + [
                    get_erc20_contract(self.w3, token_address).functions.decimals()
                    for token_address in current_tokens
                ]
            )
            token_balances = token_balances_and_decimals[:number_tokens]
            token_decimals = token_balances_and_decimals[number_tokens:]
            token_prices = [
                self.price_oracle.get_price(token_address)
                for token_address in current_tokens
            ]
            total_eth_value = 0
            for token_balance, token_decimal, token_price in zip(
                token_balances, token_decimals, token_prices
            ):
                total_eth_value += (token_balance / 10**token_decimal) * token_price
            return total_eth_value / (total_supply / 1e18)
        except (Web3Exception, DecodingError, ValueError):
            raise CannotGetPriceFromOracle(
                f"Cannot get price for {pool_token_address}. "
                f"It is not a balancer pool token"
            )


class MooniswapOracle(BalancerOracle):
    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate balancer pool token price based on its components

        :param pool_token_address: Moniswap pool token address
        :return: Eth price for pool token
        :raises: CannotGetPriceFromOracle
        """
        try:
            balancer_pool_contract = self.w3.eth.contract(
                pool_token_address, abi=mooniswap_abi
            )
            tokens, total_supply = self.ethereum_client.batch_call(
                [
                    balancer_pool_contract.functions.getTokens(),
                    balancer_pool_contract.functions.totalSupply(),
                ]
            )
            if not tokens:
                raise ValueError
            if len(tokens) == 1 or any(
                [token == NULL_ADDRESS for token in tokens]
            ):  # One of the tokens is ether
                ethereum_amount = self.ethereum_client.get_balance(pool_token_address)
                return ethereum_amount * 2 / total_supply
            else:
                for token in tokens:
                    try:
                        price = self.price_oracle.get_price(token)
                        token_contract = get_erc20_contract(self.w3, token)
                        token_balance, token_decimals = self.ethereum_client.batch_call(
                            [
                                token_contract.functions.balanceOf(pool_token_address),
                                token_contract.functions.decimals(),
                            ]
                        )
                        total_value = (token_balance / 10**token_decimals) * price
                        return (total_value * 2) / (total_supply / 1e18)
                    except CannotGetPriceFromOracle:
                        continue

                raise CannotGetPriceFromOracle(
                    f"Cannot get price for {pool_token_address}. "
                    f"It is not a mooniswap pool token"
                )

        except (Web3Exception, DecodingError, ValueError):
            raise CannotGetPriceFromOracle(
                f"Cannot get price for {pool_token_address}. "
                f"It is not a mooniswap pool token"
            )



import os

from django.test import TestCase

from eth_abi.packed import encode_packed
from eth_utils import to_checksum_address
from hexbytes import HexBytes

from ..contracts import get_proxy_1_0_0_deployed_bytecode, get_proxy_factory_contract
from ..utils import (
    compare_byte_code,
    decode_string_or_bytes32,
    fast_bytes_to_checksum_address,
    fast_is_checksum_address,
    fast_keccak,
    fast_to_checksum_address,
    mk_contract_address,
    mk_contract_address_2,
)
from .ethereum_test_case import EthereumTestCaseMixin


class TestUtils(EthereumTestCaseMixin, TestCase):
    def test_mk_contract_address_2(self):
        from_ = "0x8942595A2dC5181Df0465AF0D7be08c8f23C93af"
        salt = self.w3.keccak(text="aloha")
        init_code = "0x00abcd"
        expected = "0x8D02C796Dd019916F65EBa1C9D65a7079Ece00E0"
        address2 = mk_contract_address_2(from_, salt, init_code)
        self.assertEqual(address2, expected)

        from_ = HexBytes("0x8942595A2dC5181Df0465AF0D7be08c8f23C93af")
        salt = self.w3.keccak(text="aloha").hex()
        init_code = HexBytes("0x00abcd")
        expected = "0x8D02C796Dd019916F65EBa1C9D65a7079Ece00E0"
        address2 = mk_contract_address_2(from_, salt, init_code)
        self.assertEqual(address2, expected)

    def test_generate_address2_with_proxy(self):
        deployer_account = self.ethereum_test_account
        proxy_factory_contract = get_proxy_factory_contract(self.w3)
        nonce = self.w3.eth.get_transaction_count(
            deployer_account.address, block_identifier="pending"
        )
        tx = proxy_factory_contract.constructor().build_transaction(
            {"nonce": nonce, "from": deployer_account.address}
        )
        signed_tx = deployer_account.sign_transaction(tx)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        proxy_factory_contract = get_proxy_factory_contract(
            self.w3, address=tx_receipt["contractAddress"]
        )

        initializer = b""  # Should be the safe `setup()` call with `owners`, `threshold`, `payment`...
        salt_nonce = 0  # Random. For sure. I used a dice

        # From v1.4.1 createProxyWithNonce requires a valid singleton address, so we will use any deployed contract
        master_copy = proxy_factory_contract.address
        tx = proxy_factory_contract.functions.createProxyWithNonce(
            master_copy, initializer, salt_nonce
        ).build_transaction(
            {
                "nonce": nonce + 1,
                "from": deployer_account.address,
            }
        )
        signed_tx = deployer_account.sign_transaction(tx)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        logs = proxy_factory_contract.events.ProxyCreation().process_receipt(tx_receipt)
        log = logs[0]
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]

        proxy_creation_code = (
            proxy_factory_contract.functions.proxyCreationCode().call()
        )
        salt = self.w3.keccak(
            encode_packed(
                ["bytes", "uint256"], [self.w3.keccak(initializer), salt_nonce]
            )
        )
        deployment_data = encode_packed(
            ["bytes", "uint256"], [proxy_creation_code, int(master_copy, 16)]
        )
        address2 = mk_contract_address_2(
            proxy_factory_contract.address, salt, deployment_data
        )
        self.assertEqual(proxy_address, address2)

    def test_decode_string_or_bytes32(self):
        # Abi encoded string
        gnosis_hex = HexBytes(
            "0x00000000000000000000000000000000000000000000000000000000000000200000000000000000000000"
            "00000000000000000000000000000000000000000c476e6f73697320546f6b656e0000000000000000000000"
            "000000000000000000"
        )

        # Abi encoded bytes32
        dai_hex = HexBytes(
            "0x44616920537461626c65636f696e2076312e3000000000000000000000000000"
        )

        self.assertEqual(decode_string_or_bytes32(gnosis_hex), "Gnosis Token")
        self.assertEqual(decode_string_or_bytes32(dai_hex), "Dai Stablecoin v1.0")

    def test_compare_byte_code(self):
        proxy_with_metadata = get_proxy_1_0_0_deployed_bytecode()
        proxy_with_different_metadata = HexBytes(
            "0x608060405273ffffffffffffffffffffffffffffffffffffffff60005416366000"
            "8037600080366000845af43d6000803e6000811415603d573d6000fd5b3d6000f3fe"
            "a165627a7a72305820b7f4e514a2bdfeb2e729e84f7101233a43b51d677007041e70"
            "8067e4b88bec480029"
        )
        self.assertTrue(
            compare_byte_code(proxy_with_metadata, proxy_with_different_metadata)
        )

    def test_fast_keccak(self):
        text = "chidori"
        self.assertEqual(
            fast_keccak(text.encode()),
            HexBytes(
                "0xd20148e42186a9e7698e652e255c809851633d62bad625a55d05efd7f718449c"
            ),
        )

        binary = b""
        self.assertEqual(
            fast_keccak(binary),
            HexBytes(
                "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"
            ),
        )

        binary = b"1234"
        self.assertEqual(
            fast_keccak(binary),
            HexBytes(
                "0x387a8233c96e1fc0ad5e284353276177af2186e7afa85296f106336e376669f7"
            ),
        )

    def test_fast_to_checksum_address(self):
        for _ in range(10):
            address = os.urandom(20).hex()
            self.assertEqual(
                fast_to_checksum_address(address), to_checksum_address(address)
            )

    def test_fast_bytes_to_checksum_address(self):
        with self.assertRaises(ValueError):
            fast_bytes_to_checksum_address(os.urandom(19))

        with self.assertRaises(ValueError):
            fast_bytes_to_checksum_address(os.urandom(21))

        for _ in range(10):
            address = os.urandom(20)
            self.assertEqual(
                fast_bytes_to_checksum_address(address), to_checksum_address(address)
            )

    def test_fast_is_checksum_address(self):
        self.assertFalse(fast_is_checksum_address(None))
        self.assertFalse(fast_is_checksum_address(""))
        self.assertFalse(
            fast_is_checksum_address(0x6ED857DC1DA2C41470A95589BB482152000773E9)
        )
        self.assertFalse(fast_is_checksum_address(2))
        self.assertFalse(fast_is_checksum_address("2"))
        self.assertFalse(
            fast_is_checksum_address("0x6ed857dc1da2c41470A95589bB482152000773e9")
        )
        self.assertTrue(
            fast_is_checksum_address("0x6ED857dc1da2c41470A95589bB482152000773e9")
        )

    def test_mk_contract_address(self):
        self.assertEqual(
            mk_contract_address("0x76E2cFc1F5Fa8F6a5b3fC4c8F4788F0116861F9B", 129),
            "0x5A317285cD83092fD40153C4B7c3Df64d8482Da8",
        )
        self.assertEqual(
            mk_contract_address("0x76E2cFc1F5Fa8F6a5b3fC4c8F4788F0116861F9B", 399),
            "0xF03b503CC9Ee8aAA3B17856942a440be0c77Cd84",
        )



from eth.constants import SECPK1_N
from eth_typing import ChecksumAddress, HexAddress, HexStr

NULL_ADDRESS: ChecksumAddress = ChecksumAddress(HexAddress(HexStr("0x" + "0" * 40)))
SENTINEL_ADDRESS: ChecksumAddress = ChecksumAddress(
    HexAddress(HexStr("0x" + "0" * 39 + "1"))
)

SAFE_SINGLETON_FACTORY_DEPLOYER_ADDRESS: ChecksumAddress = ChecksumAddress(
    HexAddress(HexStr("0xE1CB04A0fA36DdD16a06ea828007E35e1a3cBC37"))
)
SAFE_SINGLETON_FACTORY_ADDRESS: ChecksumAddress = ChecksumAddress(
    HexAddress(HexStr("0x914d7Fec6aaC8cd542e72Bca78B30650d45643d7"))
)

# keccak('Transfer(address,address,uint256)')
ERC20_721_TRANSFER_TOPIC: HexStr = HexStr(
    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
)

SIGNATURE_R_MIN_VALUE: int = 1
SIGNATURE_R_MAX_VALUE: int = SECPK1_N - 1
SIGNATURE_S_MIN_VALUE: int = 1
SIGNATURE_S_MAX_VALUE: int = SECPK1_N // 2
SIGNATURE_V_MIN_VALUE: int = 27
SIGNATURE_V_MAX_VALUE: int = 28


GAS_CALL_DATA_ZERO_BYTE = 4
GAS_CALL_DATA_BYTE = 16  # 68 before Istanbul



import copy
from unittest import mock
from unittest.mock import MagicMock, PropertyMock, patch

from django.test import TestCase

from eth_account import Account
from eth_typing import HexStr
from hexbytes import HexBytes

from gnosis.eth import EthereumClient, EthereumNetwork, EthereumNetworkNotSupported
from gnosis.eth.tests.ethereum_test_case import EthereumTestCaseMixin
from gnosis.safe import SafeTx
from gnosis.safe.api.transaction_service_api import (
    ApiSafeTxHashNotMatchingException,
    TransactionServiceApi,
)

from ...api import SafeAPIException
from ..mocks.mock_transactions import (
    transaction_data_decoded_mock,
    transaction_data_mock,
    transaction_mock,
)


class TestTransactionServiceAPI(EthereumTestCaseMixin, TestCase):
    def setUp(self) -> None:
        self.transaction_service_api = TransactionServiceApi(
            EthereumNetwork.GNOSIS, ethereum_client=self.ethereum_client
        )
        self.safe_address = "0xAedF684C1c41B51CbD228116e11484425d2FACB9"

    def test_constructor(self):
        ethereum_network = EthereumNetwork.GOERLI
        base_url = "https://safe.global"
        transaction_service_api = TransactionServiceApi(
            ethereum_network, ethereum_client=None, base_url=base_url
        )
        self.assertEqual(transaction_service_api.network, ethereum_network)
        self.assertIsNone(transaction_service_api.ethereum_client)
        self.assertEqual(transaction_service_api.base_url, base_url)

    def test_from_ethereum_client(self):
        with self.assertRaisesMessage(EthereumNetworkNotSupported, "GANACHE"):
            TransactionServiceApi.from_ethereum_client(self.ethereum_client)

        with mock.patch.object(
            EthereumClient, "get_network", return_value=EthereumNetwork.GOERLI
        ):
            transaction_service_api = TransactionServiceApi.from_ethereum_client(
                self.ethereum_client
            )
            self.assertEqual(
                transaction_service_api.ethereum_client, self.ethereum_client
            )
            self.assertEqual(transaction_service_api.network, EthereumNetwork.GOERLI)

    def test_data_decoded_to_text(self):
        decoded_data_text = self.transaction_service_api.data_decoded_to_text(
            transaction_data_mock
        )
        self.assertIn(
            "- changeMasterCopy: 0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F",
            decoded_data_text,
        )
        self.assertIn(
            "- setFallbackHandler: 0xd5D82B6aDDc9027B22dCA772Aa68D5d74cdBdF44",
            decoded_data_text,
        )

    def test_get_balances(self):
        balances = self.transaction_service_api.get_balances(self.safe_address)
        self.assertIsInstance(balances, list)
        self.assertGreaterEqual(len(balances), 1)

    def test_get_transactions(self):
        transactions = self.transaction_service_api.get_transactions(self.safe_address)
        self.assertIsInstance(transactions, list)
        self.assertEqual(len(transactions), 6)

        # Test arguments
        with patch.object(TransactionServiceApi, "_get_request") as mock_get_request:
            self.transaction_service_api.get_transactions(
                self.safe_address, limit=2, nonce__lt=30, failed=False
            )

        expected_url = f"/api/v1/safes/{self.safe_address}/multisig-transactions/?limit=2&nonce__lt=30&failed=False"
        mock_get_request.assert_called_once_with(expected_url)

        # Test valid safe tx has
        with patch.object(
            SafeTx, "safe_version", new_callable=PropertyMock
        ) as mock_safe_version:
            mock_safe_version.return_value = "1.4.1"
            safe_tx_hash = (
                "0x06c88df42a8ab64b2b2c5e2b5c8c4df384c267b39929a8416d1518db23f91783"
            )
            transactions = self.transaction_service_api.get_transactions(
                self.safe_address, safe_tx_hash=safe_tx_hash
            )
            self.assertIsInstance(transactions, list)
            self.assertEqual(len(transactions), 1)

        # Test invalid safe tx hash
        with (
            patch.object(TransactionServiceApi, "_get_request") as mock_get_request,
            patch.object(
                SafeTx, "safe_version", new_callable=PropertyMock
            ) as mock_safe_version,
        ):
            mock_safe_version.return_value = "1.4.1"
            mock_get_request.return_value.ok = True
            mock_get_request.return_value.json = MagicMock(
                return_value={"results": [transaction_mock]}
            )

            safe_tx_invalid_hash = (
                "0x06c88df42a8ab64b2b2c5e2b5c8c4df384c267b39929a8416d1518db23f90000"
            )
            with self.assertRaises(ApiSafeTxHashNotMatchingException) as context:
                self.transaction_service_api.get_transactions(
                    self.safe_address, safe_tx_hash=safe_tx_invalid_hash
                )
            safe_tx_hash_expected = transaction_mock.get("safeTxHash")
            self.assertIn(
                f"API safe-tx-hash: {safe_tx_invalid_hash} doesn't match the calculated safe-tx-hash: {safe_tx_hash_expected}",
                str(context.exception),
            )

    def test_get_safes_for_owner(self):
        owner_address = "0x3066786706Ff0B6e71044e55074dBAE7D01573cB"
        safes = self.transaction_service_api.get_safes_for_owner(owner_address)
        self.assertListEqual(safes, [self.safe_address])

    def test_get_safe_transaction(self):
        safe_tx_hash = HexBytes(
            "0x06c88df42a8ab64b2b2c5e2b5c8c4df384c267b39929a8416d1518db23f91783"
        )
        with (
            patch.object(TransactionServiceApi, "_get_request") as mock_get_request,
            patch.object(
                SafeTx, "safe_version", new_callable=PropertyMock
            ) as mock_safe_version,
        ):
            mock_safe_version.return_value = "1.4.1"
            mock_get_request.return_value.ok = True
            mock_get_request.return_value.json = MagicMock(
                return_value=transaction_mock
            )

            safe_tx, tx_hash = self.transaction_service_api.get_safe_transaction(
                safe_tx_hash
            )
            self.assertEqual(safe_tx.tx_hash, tx_hash)
            self.assertEqual(safe_tx.safe_version, "1.4.1")
            self.assertEqual(safe_tx.safe_tx_hash, safe_tx_hash)

            # Test not executed tx
            not_executed_transaction_mock = copy.deepcopy(transaction_mock)
            not_executed_transaction_mock["transactionHash"] = None
            mock_get_request.return_value.json = MagicMock(
                return_value=not_executed_transaction_mock
            )
            safe_tx, tx_hash = self.transaction_service_api.get_safe_transaction(
                safe_tx_hash
            )
            self.assertIsNone(tx_hash)
            self.assertIsNone(safe_tx.tx_hash)

            # Test invalid safe tx hash
            safe_tx_invalid_hash = HexBytes(
                "0x06c88df42a8ab64b2b2c5e2b5c8c4df384c267b39929a8416d1518db23f90000"
            )
            with self.assertRaises(ApiSafeTxHashNotMatchingException) as context:
                self.transaction_service_api.get_safe_transaction(safe_tx_invalid_hash)

            safe_tx_hash_expected = transaction_mock.get("safeTxHash")
            self.assertIn(
                f"API safe-tx-hash: {safe_tx_invalid_hash.hex()} doesn't match the calculated safe-tx-hash: {safe_tx_hash_expected}",
                str(context.exception),
            )

            # Test response not ok
            mock_get_request.return_value.ok = False
            with self.assertRaises(SafeAPIException) as context:
                self.transaction_service_api.get_safe_transaction(safe_tx_hash)
            self.assertIn(
                f"Cannot get transaction with safe-tx-hash={safe_tx_hash.hex()}:",
                str(context.exception),
            )

    def test_remove_delegate(self):
        with patch.object(
            TransactionServiceApi, "_delete_request"
        ) as mock_delete_request:
            delegate_address = Account().create().address
            delegator_account = Account().create()
            message_hash = self.transaction_service_api.create_delegate_message_hash(
                delegate_address
            )
            signature = delegator_account.signHash(message_hash).signature.hex()
            self.transaction_service_api.remove_delegate(
                delegate_address, delegator_account.address, signature
            )

            expected_url = f"/api/v2/delegates/{delegate_address}/"
            expected_payload = {
                "delegator": delegator_account.address,
                "signature": signature,
            }
            mock_delete_request.assert_called_once_with(expected_url, expected_payload)

            self.transaction_service_api.remove_delegate(
                delegate_address,
                delegator_account.address,
                signature,
                self.safe_address,
            )

            expected_payload = {
                "safe": self.safe_address,
                "delegator": delegator_account.address,
                "signature": signature,
            }
            mock_delete_request.assert_called_with(expected_url, expected_payload)

    def test_add_delegate(self):
        with patch.object(TransactionServiceApi, "_post_request") as mock_post_request:
            delegate_address = Account().create().address
            delegator_account = Account().create()
            message_hash = self.transaction_service_api.create_delegate_message_hash(
                delegate_address
            )
            signature = delegator_account.signHash(message_hash).signature.hex()
            label = "test label"
            self.transaction_service_api.add_delegate(
                delegate_address, delegator_account.address, label, signature
            )

            expected_url = "/api/v2/delegates/"
            expected_payload = {
                "delegate": delegate_address,
                "delegator": delegator_account.address,
                "signature": signature,
                "label": label,
            }
            mock_post_request.assert_called_once_with(expected_url, expected_payload)

            self.transaction_service_api.add_delegate(
                delegate_address,
                delegator_account.address,
                label,
                signature,
                self.safe_address,
            )

            expected_payload = {
                "safe": self.safe_address,
                "delegate": delegate_address,
                "delegator": delegator_account.address,
                "signature": signature,
                "label": label,
            }
            mock_post_request.assert_called_with(expected_url, expected_payload)

    def test_decode_data(self):
        with patch.object(TransactionServiceApi, "_post_request") as mock_post_request:
            mock_post_request.return_value.ok = True
            mock_post_request.return_value.json = MagicMock(
                return_value=transaction_data_decoded_mock
            )
            data = HexStr(
                "0x095ea7b3000000000000000000000000e6fc577e87f7c977c4393300417dcc592d90acf8ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
            )
            to_address = "0x4127839cdf4F73d9fC9a2C2861d8d1799e9DF40C"
            decoded_data = self.transaction_service_api.decode_data(data, to_address)
            expected_url = "/api/v1/data-decoder/"
            expected_payload = {"data": HexBytes(data).hex(), "to": to_address}
            mock_post_request.assert_called_once_with(expected_url, expected_payload)
            self.assertEqual(decoded_data.get("method"), "approve")
            self.assertEqual(decoded_data.get("parameters")[0].get("name"), "spender")
            self.assertEqual(decoded_data.get("parameters")[1].get("name"), "value")

            # Test response not ok
            mock_post_request.return_value.ok = False
            with self.assertRaises(SafeAPIException) as context:
                self.transaction_service_api.decode_data(data, to_address)
            self.assertIn(
                "Cannot decode tx data:",
                str(context.exception),
            )



from django.contrib.admin.utils import lookup_spawns_duplicates
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.utils.text import smart_split, unescape_string_literal


class AdvancedAdminSearchMixin:
    """
    Use database indexes when using exact search instead
    of converting everything to text before searching
    """

    def get_search_results(self, request, queryset, search_term):
        """
        Return a tuple containing a queryset to implement the search
        and a boolean indicating if the results may contain duplicates.

        This function was modified from Django original get_search_results
        to allow `exact` search that uses database indexes
        """

        def construct_search(field_name):
            if field_name.startswith("^"):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith("=="):
                return "%s__exact" % field_name[2:]
            elif field_name.startswith("="):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith("@"):
                return "%s__search" % field_name[1:]
            # Use field_name if it includes a lookup.
            opts = queryset.model._meta
            lookup_fields = field_name.split(LOOKUP_SEP)
            # Go through the fields, following all relations.
            prev_field = None
            for path_part in lookup_fields:
                if path_part == "pk":
                    path_part = opts.pk.name
                try:
                    field = opts.get_field(path_part)
                except FieldDoesNotExist:
                    # Use valid query lookups.
                    if prev_field and prev_field.get_lookup(path_part):
                        return field_name
                else:
                    prev_field = field
                    if hasattr(field, "path_infos"):
                        # Update opts to follow the relation.
                        opts = field.path_infos[-1].to_opts
            # Otherwise, use the field with icontains.
            return "%s__icontains" % field_name

        may_have_duplicates = False
        search_fields = self.get_search_fields(request)
        if search_fields and search_term:
            orm_lookups = [
                construct_search(str(search_field)) for search_field in search_fields
            ]
            term_queries = []
            for bit in smart_split(search_term):
                if bit.startswith(('"', "'")) and bit[0] == bit[-1]:
                    bit = unescape_string_literal(bit)

                valid_queries = []
                for orm_lookup in orm_lookups:
                    try:
                        # Check if query is valid (for example, not a number provided for an integer exact query)
                        # This is the main difference comparing to Django official implementation
                        queryset.filter(**{orm_lookup: bit})
                        valid_queries.append((orm_lookup, bit))
                    except (ValueError, ValidationError):
                        pass
                or_queries = models.Q.create(
                    [valid_query for valid_query in valid_queries],
                    connector=models.Q.OR,
                )
                term_queries.append(or_queries)
            queryset = queryset.filter(models.Q.create(term_queries))
            may_have_duplicates |= any(
                lookup_spawns_duplicates(self.opts, search_spec)
                for search_spec in orm_lookups
            )
        return queryset, may_have_duplicates



from unittest import mock

from django.test import TestCase

import pytest
from eth_account import Account
from requests import Session

from ... import EthereumClient
from ...oracles import CannotGetPriceFromOracle, CowswapOracle
from ..ethereum_test_case import EthereumTestCaseMixin
from ..test_oracles import (
    dai_token_mainnet_address,
    usdc_token_mainnet_address,
    usdt_token_mainnet_address,
    weth_token_mainnet_address,
)
from ..utils import just_test_if_mainnet_node


@pytest.skip("Having issues often", allow_module_level=True)
class TestCowswapOracle(EthereumTestCaseMixin, TestCase):
    def test_get_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(CowswapOracle.is_available(ethereum_client))

        cowswap_oracle = CowswapOracle(ethereum_client)

        price = cowswap_oracle.get_price(weth_token_mainnet_address)
        self.assertEqual(price, 1.0)

        price = cowswap_oracle.get_price(dai_token_mainnet_address)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        price = cowswap_oracle.get_price(
            weth_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertGreater(price, 1)

        # Test with 2 stablecoins with same decimals
        price = cowswap_oracle.get_price(
            usdt_token_mainnet_address, usdc_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        # Test with 2 stablecoins with different decimals
        price = cowswap_oracle.get_price(
            dai_token_mainnet_address, usdc_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        price = cowswap_oracle.get_price(
            usdc_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        with mock.patch.object(
            Session, "post", side_effect=IOError("Connection Error")
        ):
            with self.assertRaises(CannotGetPriceFromOracle):
                cowswap_oracle.get_price(
                    usdc_token_mainnet_address, dai_token_mainnet_address
                )

        random_token = Account.create().address
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle,
            f"Cannot get decimals for token={random_token}",
        ):
            cowswap_oracle.get_price(random_token)

        with mock.patch(
            "gnosis.eth.oracles.cowswap.get_decimals", autospec=True, return_value=18
        ):
            with self.assertRaises(CannotGetPriceFromOracle):
                cowswap_oracle.get_price(random_token)



from .. import Safe
from .test_safe import TestSafe as BaseTestSafe  # So it's not run twice by pytest


class TestSafeV130(BaseTestSafe):
    @property
    def safe_contract(self):
        """
        :return: Last Safe Contract available
        """
        return self.safe_contract_V1_3_0

    def deploy_test_safe(self, *args, **kwargs):
        return self.deploy_test_safe_v1_3_0(*args, **kwargs)

    def test_retrieve_modules_unitialized_safe(self):
        """
         An unitialized V1.3.0 Safe will return `[[], '0x0000000000000000000000000000000000000000']` when calling
        `getModulesPaginated`, as `SENTINEL_ADDRESS` is only set when initialized
        """

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            self.safe_contract.address,
            initializer=b"",
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(safe.retrieve_modules(), [])
        self.assertEqual(safe.retrieve_all_info().modules, [])



"""
Based on https://github.com/jvinet/eip712, adjustments by https://github.com/uxio0

Routines for EIP712 encoding and signing.

Copyright (C) 2022 Judd Vinet <jvinet@zeroflux.org>
                   Uxío Fuentefría <uxio@safe.global>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
from typing import Any, Dict, Tuple

from eth_abi import encode as encode_abi
from eth_typing import Hash32
from hexbytes import HexBytes

from ..utils import fast_keccak


def encode_data(primary_type: str, data, types):
    """
    Encode structured data as per Ethereum's signTypeData_v4.

    https://docs.metamask.io/guide/signing-data.html#sign-typed-data-v4

    This code is ported from the Javascript "eth-sig-util" package.
    """
    encoded_types = ["bytes32"]
    encoded_values = [hash_type(primary_type, types)]

    def _encode_field(name, typ, value):
        if typ in types:
            if value is None:
                return [
                    "bytes32",
                    "0x0000000000000000000000000000000000000000000000000000000000000000",
                ]
            else:
                return ["bytes32", fast_keccak(encode_data(typ, value, types))]

        if value is None:
            raise Exception(f"Missing value for field {name} of type {type}")

        # Accept string bytes
        if "bytes" in typ and isinstance(value, str):
            value = HexBytes(value)

        # Accept string uint and int
        if "int" in typ and isinstance(value, str):
            value = int(value)

        if typ == "bytes":
            return ["bytes32", fast_keccak(value)]

        if typ == "string":
            # Convert string to bytes.
            value = value.encode("utf-8")
            return ["bytes32", fast_keccak(value)]

        if typ.endswith("]"):
            # Array type
            if value:
                parsed_type = typ[: typ.rindex("[")]
                type_value_pairs = [_encode_field(name, parsed_type, v) for v in value]
                data_types, data_hashes = zip(*type_value_pairs)
            else:
                # Empty array
                data_types, data_hashes = [], []

            h = fast_keccak(encode_abi(data_types, data_hashes))
            return ["bytes32", h]

        return [typ, value]

    for field in types[primary_type]:
        typ, val = _encode_field(field["name"], field["type"], data[field["name"]])
        encoded_types.append(typ)
        encoded_values.append(val)

    return encode_abi(encoded_types, encoded_values)


def encode_type(primary_type: str, types) -> str:
    result = ""
    deps = find_type_dependencies(primary_type, types)
    deps = sorted([d for d in deps if d != primary_type])
    deps = [primary_type] + deps
    for typ in deps:
        children = types[typ]
        if not children:
            raise Exception(f"No type definition specified: {type}")

        defs = [f"{t['type']} {t['name']}" for t in types[typ]]
        result += typ + "(" + ",".join(defs) + ")"
    return result


def find_type_dependencies(primary_type: str, types, results=None):
    if results is None:
        results = []

    primary_type = re.split(r"\W", primary_type)[0]
    if primary_type in results or not types.get(primary_type):
        return results
    results.append(primary_type)

    for field in types[primary_type]:
        deps = find_type_dependencies(field["type"], types, results)
        for dep in deps:
            if dep not in results:
                results.append(dep)

    return results


def hash_type(primary_type: str, types) -> Hash32:
    return fast_keccak(encode_type(primary_type, types).encode())


def hash_struct(primary_type: str, data, types) -> Hash32:
    return fast_keccak(encode_data(primary_type, data, types))


def eip712_encode(typed_data: Dict[str, Any]) -> Tuple[bytes, Hash32, Hash32]:
    """
    Given a dict of structured data and types, return a 3-element tuple of
    the encoded, signable data.

      0: The magic & version (0x1901)
      1: The encoded types
      2: The encoded data
    """
    try:
        parts = (
            bytes.fromhex("1901"),
            hash_struct("EIP712Domain", typed_data["domain"], typed_data["types"]),
        )
        if typed_data["primaryType"] == "EIP712Domain":
            raise TypeError("primaryType cannot be EIP712Domain")

        return (
            *parts,
            hash_struct(
                typed_data["primaryType"],
                typed_data["message"],
                typed_data["types"],
            ),
        )

    except (KeyError, AttributeError, TypeError, IndexError) as exc:
        raise ValueError(f"Not valid {typed_data}") from exc


def eip712_encode_hash(typed_data: Dict[str, Any]) -> Hash32:
    """
    :param typed_data: EIP712 structured data and types
    :return: Keccak256 hash of encoded signable data
    """
    return fast_keccak(b"".join(eip712_encode(typed_data)))



from django.core.exceptions import ValidationError
from django.core.serializers import serialize
from django.db import DataError, transaction
from django.test import TestCase

from eth_account import Account
from faker import Faker

from ...constants import NULL_ADDRESS, SENTINEL_ADDRESS
from ...utils import fast_is_checksum_address, fast_keccak_text
from .models import (
    EthereumAddress,
    EthereumAddressV2,
    Keccak256Hash,
    Sha3Hash,
    Uint32,
    Uint96,
    Uint256,
)

faker = Faker()


class TestModels(TestCase):
    def test_ethereum_address_field(self):
        for EthereumAddressModel in (EthereumAddress, EthereumAddressV2):
            with self.subTest(EthereumAddressModel=EthereumAddressModel):
                address = Account.create().address
                self.assertTrue(fast_is_checksum_address(address))
                ethereum_address = EthereumAddressModel.objects.create(value=address)
                ethereum_address.refresh_from_db()
                self.assertTrue(fast_is_checksum_address(ethereum_address.value))
                self.assertEqual(address, ethereum_address.value)

                # Test addresses
                for addresss in (
                    None,
                    NULL_ADDRESS,
                    SENTINEL_ADDRESS,
                    Account.create().address,
                ):
                    with self.subTest(special_address=addresss):
                        EthereumAddressModel.objects.create(value=addresss)
                        self.assertEqual(
                            EthereumAddressModel.objects.get(value=addresss).value,
                            addresss,
                        )

                with self.assertRaisesMessage(
                    ValidationError,
                    '"0x23" value must be an EIP55 checksummed address.',
                ):
                    with transaction.atomic():
                        EthereumAddressModel.objects.create(value="0x23")

                ethereum_address = EthereumAddressModel(value=Account.create().address)
                self.assertIsNone(ethereum_address.full_clean())

    def test_uint256_field(self):
        for value in [
            2,
            2**256,
            25572735541615049941137326092682691158109824779649981270427004917341670006487,
            None,
        ]:
            uint256 = Uint256.objects.create(value=value)
            uint256.refresh_from_db()
            self.assertEqual(uint256.value, value)

        # Overflow
        with self.assertRaises(Exception):
            value = 2**263
            Uint256.objects.create(value=value)

        # Signed
        with self.assertRaises(ValidationError):
            Uint256.objects.create(value=-2)

    def test_uint96_field(self):
        for value in [
            2,
            2**96,
            79228162514264337593543950336,
            None,
        ]:
            uint96 = Uint96.objects.create(value=value)
            uint96.refresh_from_db()
            self.assertEqual(uint96.value, value)

        # Overflow
        with self.assertRaises(Exception):
            value = 2**97
            Uint96.objects.create(value=value)

        # Signed
        with self.assertRaises(ValidationError):
            Uint96.objects.create(value=-2)

    def test_uint32_field(self):
        for value in [
            2,
            2**32,
            4294967296,
            None,
        ]:
            uint32 = Uint32.objects.create(value=value)
            uint32.refresh_from_db()
            self.assertEqual(uint32.value, value)
        # Overflow
        with self.assertRaises(Exception):
            value = 2**34
            Uint32.objects.create(value=value)

        # Signed
        with self.assertRaises(ValidationError):
            Uint32.objects.create(value=-2)

    def test_sha3_hash_field(self):
        value_hexbytes = fast_keccak_text(faker.name())
        value_hex_with_0x: str = value_hexbytes.hex()
        value_hex_without_0x: str = value_hex_with_0x[2:]
        value: bytes = bytes(value_hexbytes)

        values = [value, value_hex_without_0x, value_hex_with_0x, value_hexbytes]

        for v in values:
            sha3_hash = Sha3Hash.objects.create(value=v)
            sha3_hash.refresh_from_db()
            self.assertEqual(sha3_hash.value, value_hex_with_0x)

        for v in values:
            self.assertEqual(Sha3Hash.objects.filter(value=v).count(), len(values))

        # Hash null
        sha3_hash = Sha3Hash.objects.create(value=None)
        sha3_hash.refresh_from_db()
        self.assertIsNone(sha3_hash.value)

        # Hash too big
        value_hex_invalid: str = "0x" + value_hex_without_0x + "a"
        with self.assertRaisesMessage(
            DataError, "value too long for type character varying(64)"
        ):
            with transaction.atomic():
                Sha3Hash.objects.create(value=value_hex_invalid)

    def test_keccak256_field(self):
        value_hexbytes = fast_keccak_text(faker.name())
        value_hex_with_0x: str = value_hexbytes.hex()
        value_hex_without_0x: str = value_hex_with_0x[2:]
        value: bytes = bytes(value_hexbytes)

        values = [value, value_hex_without_0x, value_hex_with_0x, value_hexbytes]

        for v in values:
            with self.subTest(v=v):
                keccak256_hash = Keccak256Hash(value=v)
                self.assertIsNone(keccak256_hash.full_clean())
                keccak256_hash.save()
                keccak256_hash.refresh_from_db()
                self.assertEqual(keccak256_hash.value, value_hex_with_0x)

        for v in values:
            with self.subTest(v=v):
                self.assertEqual(
                    Keccak256Hash.objects.filter(value=v).count(), len(values)
                )

        # Hash null
        keccak256_hash = Keccak256Hash.objects.create(value=None)
        keccak256_hash.refresh_from_db()
        self.assertIsNone(keccak256_hash.value)

        # Hash too big
        value_hex_invalid: str = "0x" + value_hex_without_0x + "a"
        with self.assertRaisesMessage(
            ValidationError, f'"{value_hex_invalid}" hash must have exactly 32 bytes.'
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)

        # Hash too small
        value_hex_invalid: str = "0x" + "a1"
        with self.assertRaisesMessage(
            ValidationError, f'"{value_hex_invalid}" hash must have exactly 32 bytes.'
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)

        # Invalid hash
        value_hex_invalid: str = "UX/IO"
        with self.assertRaisesMessage(
            ValidationError,
            f'"{value_hex_invalid}" hash must be a 32 bytes hexadecimal.',
        ):
            with transaction.atomic():
                Keccak256Hash.objects.create(value=value_hex_invalid)

    def test_serialize_keccak256_field_to_json(self):
        hexvalue: str = (
            "0xdb5b7c6d3b0cc538a5859afc4674a785d9d111c3835390295f3d3173d41ca8ea"
        )
        Keccak256Hash.objects.create(value=hexvalue)
        serialized = serialize("json", Keccak256Hash.objects.all())
        # hexvalue should be in serialized data
        self.assertIn(hexvalue, serialized)

    def test_serialize_ethereum_address_field_to_json(self):
        address: str = "0x5aFE3855358E112B5647B952709E6165e1c1eEEe"
        EthereumAddress.objects.create(value=address)
        serialized = serialize("json", EthereumAddress.objects.all())
        # address should be in serialized data
        self.assertIn(address, serialized)

    def test_serialize_ethereum_address_v2_field_to_json(self):
        address: str = "0x5aFE3855358E112B5647B952709E6165e1c1eEEe"
        EthereumAddressV2.objects.create(value=address)
        serialized = serialize("json", EthereumAddressV2.objects.all())
        # address should be in serialized data
        self.assertIn(address, serialized)

    def test_serialize_uint256_field_to_json(self):
        value = 2**256
        Uint256.objects.create(value=value)
        serialized = serialize("json", Uint256.objects.all())
        # value should be in serialized data
        self.assertIn(str(value), serialized)

    def test_serialize_sha3_hash_to_json(self):
        hash = fast_keccak_text("testSerializer")
        Sha3Hash.objects.create(value=hash)
        serialized = serialize("json", Sha3Hash.objects.all())
        # hash should be in serialized data
        self.assertIn(hash.hex(), serialized)



import os
from functools import cache
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from ...util.http import prepare_http_session
from .. import EthereumNetwork
from ..utils import fast_is_checksum_address
from .contract_metadata import ContractMetadata


class SourcifyClientException(Exception):
    pass


class SourcifyClientConfigurationProblem(Exception):
    pass


class SourcifyClient:
    """
    Get contract metadata from Sourcify. Matches can be full or partial:

      - Full: Both the source files as well as the meta data files were an exact match between the deployed bytecode
        and the published files.
      - Partial: Source code compiles to the same bytecode and thus the contract behaves in the same way,
        but the source code can be different: Variables can have misleading names,
        comments can be different and especially the NatSpec comments could have been modified.

    """

    def __init__(
        self,
        network: EthereumNetwork = EthereumNetwork.MAINNET,
        base_url_api: str = "https://sourcify.dev",
        base_url_repo: str = "https://repo.sourcify.dev/",
        request_timeout: int = int(
            os.environ.get("SOURCIFY_CLIENT_REQUEST_TIMEOUT", 10)
        ),
    ):
        self.network = network
        self.base_url_api = base_url_api
        self.base_url_repo = base_url_repo
        self.http_session = prepare_http_session(10, 100)
        self.request_timeout = request_timeout

        if not self.is_chain_supported(network.value):
            raise SourcifyClientConfigurationProblem(
                f"Network {network.name} - {network.value} not supported"
            )

    def _get_abi_from_metadata(self, metadata: Dict[str, Any]) -> List[Dict[str, Any]]:
        return metadata["output"]["abi"]

    def _get_name_from_metadata(self, metadata: Dict[str, Any]) -> Optional[str]:
        values = list(metadata["settings"].get("compilationTarget", {}).values())
        if values:
            return values[0]

    def _do_request(self, url: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.get(url, timeout=self.request_timeout)
        if not response.ok:
            return None

        return response.json()

    def is_chain_supported(self, chain_id: int) -> bool:
        chains = self.get_chains()
        if not chains:
            raise IOError("Cannot get chains for SourcifyClient")
        return chain_id in (int(chain["chainId"]) for chain in self.get_chains())

    @cache
    def get_chains(self) -> Dict[str, Any]:
        url = urljoin(self.base_url_api, "/server/chains")
        return self._do_request(url)

    def get_contract_metadata(
        self, contract_address: str
    ) -> Optional[ContractMetadata]:
        assert fast_is_checksum_address(
            contract_address
        ), "Expecting a checksummed address"

        for match_type in ("full_match", "partial_match"):
            url = urljoin(
                self.base_url_repo,
                f"/contracts/{match_type}/{self.network.value}/{contract_address}/metadata.json",
            )
            metadata = self._do_request(url)
            if metadata:
                abi = self._get_abi_from_metadata(metadata)
                name = self._get_name_from_metadata(metadata)
                return ContractMetadata(name, abi, match_type == "partial_match")
        return None



import logging

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes

from ...eth.utils import get_empty_tx_params
from ..enums import SafeOperationEnum
from ..exceptions import NotEnoughSafeTransactionGas, SignaturesDataTooShort
from ..multi_send import MultiSendOperation, MultiSendTx
from ..safe_tx import SafeTx
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestSafeTx(SafeTestCaseMixin, TestCase):
    def test_multi_send_safe_tx(self):
        owners = [Account.create() for _ in range(2)]
        owner_addresses = [owner.address for owner in owners]
        threshold = 1
        safe = self.deploy_test_safe(
            owners=owner_addresses,
            threshold=threshold,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_contract = safe.contract
        to = self.multi_send_contract.address
        value = 0
        safe_tx_gas = 600000
        base_gas = 200000

        # Atomic swap the owner of a Safe
        new_owner = Account.create()
        owner_to_remove = owners[-1]
        prev_owner = owners[-2]
        owners_expected = [x.address for x in owners[:-1]] + [new_owner.address]
        new_threshold = threshold + 1
        data = HexBytes(
            safe_contract.functions.addOwnerWithThreshold(
                new_owner.address, new_threshold
            ).build_transaction(get_empty_tx_params())["data"]
        )
        data_2 = HexBytes(
            safe_contract.functions.removeOwner(
                prev_owner.address, owner_to_remove.address, new_threshold
            ).build_transaction(get_empty_tx_params())["data"]
        )

        multisend_txs = [
            MultiSendTx(MultiSendOperation.CALL, safe.address, value, d)
            for d in (data, data_2)
        ]
        safe_multisend_data = self.multi_send.build_tx_data(multisend_txs)
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to,
            0,
            safe_multisend_data,
            SafeOperationEnum.DELEGATE_CALL.value,
            safe_tx_gas,
            base_gas,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        safe_tx.sign(owners[0].key)

        self.assertEqual(
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address), 1
        )
        tx_hash, _ = safe_tx.execute(
            tx_sender_private_key=self.ethereum_test_account.key
        )
        self.ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        self.assertEqual(safe.retrieve_nonce(), 1)
        self.assertEqual(safe.retrieve_threshold(), new_threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners_expected)

    def test_send_safe_tx(self):
        owners = [Account.create() for _ in range(2)]
        owner_addresses = [owner.address for owner in owners]
        threshold = 1
        safe = self.deploy_test_safe(
            owners=owner_addresses,
            threshold=threshold,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        to = Account().create().address
        value = self.w3.to_wei(0.01, "ether")
        safe_tx_gas = 200000
        data_gas = 100000

        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to,
            value,
            b"",
            0,
            safe_tx_gas,
            data_gas,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )

        with self.assertRaises(SignaturesDataTooShort):
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address)

        # Check signing
        self.assertFalse(safe_tx.signers)
        safe_tx.sign(owners[0].key)
        self.assertIn(owners[0].address, safe_tx.signers)

        with self.assertRaises(NotEnoughSafeTransactionGas):
            safe_tx.call(
                tx_sender_address=self.ethereum_test_account.address,
                tx_gas=safe_tx_gas // 2,
            )

        self.assertEqual(
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address), 1
        )
        tx_hash, _ = safe_tx.execute(
            tx_sender_private_key=self.ethereum_test_account.key
        )
        self.ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        self.assertEqual(self.ethereum_client.get_balance(to), value)

        safe_tx.unsign(owners[0].address)
        self.assertFalse(safe_tx.signers)

    def test_sign_safe_tx(self):
        owners = [Account.create() for _ in range(3)]
        owners_unsorted = sorted(owners, key=lambda x: int(x.address, 16), reverse=True)
        owner_addresses = [owner.address for owner in owners_unsorted]
        threshold = 1
        safe = self.deploy_test_safe(
            owners=owner_addresses,
            threshold=threshold,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        to = Account().create().address
        value = self.w3.to_wei(0.01, "ether")

        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to,
            value,
            b"",
            0,
            200000,
            100000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )

        safe_tx.sign(owners_unsorted[0].key)
        safe_tx.sign(owners_unsorted[2].key)
        signers = [owner_addresses[0], owner_addresses[2]]

        self.assertEqual(safe_tx.signers, safe_tx.sorted_signers)
        self.assertNotEqual(signers, safe_tx.signers)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 2)

        safe_tx.sign(owners_unsorted[1].key)
        signers = owner_addresses
        self.assertEqual(safe_tx.signers, safe_tx.sorted_signers)
        self.assertNotEqual(signers, safe_tx.signers)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 3)

        # Sign again
        safe_tx.sign(owners_unsorted[0].key)
        self.assertEqual(len(safe_tx.signers), 3)

        # Sign again
        safe_tx.unsign(owners_unsorted[1].address)
        signers = [owner_addresses[0], owner_addresses[2]]
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 2)

    def test_unsign_safe_tx(self):
        owners = [Account.create() for _ in range(3)]
        owners_unsorted = sorted(owners, key=lambda x: int(x.address, 16), reverse=True)
        owner_addresses = [owner.address for owner in owners_unsorted]
        threshold = 1
        safe = self.deploy_test_safe(
            owners=owner_addresses,
            threshold=threshold,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        to = Account().create().address
        value = self.w3.to_wei(0.01, "ether")

        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to,
            value,
            b"",
            0,
            200000,
            100000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )

        safe_tx.sign(owners_unsorted[0].key)
        safe_tx.sign(owners_unsorted[1].key)
        safe_tx.sign(owners_unsorted[2].key)
        signers = [owner_addresses[0], owner_addresses[1], owner_addresses[2]]

        self.assertEqual(safe_tx.signers, safe_tx.sorted_signers)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 3)

        # test unsign
        signature_1_removed = safe_tx.unsign(owners_unsorted[1].address)
        self.assertTrue(signature_1_removed)
        signers = [owner_addresses[0], owner_addresses[2]]
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 2)

        signature_2_removed = safe_tx.unsign(owners_unsorted[2].address)
        self.assertTrue(signature_2_removed)
        signers = [owner_addresses[0]]
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 1)

        signature_0_removed = safe_tx.unsign(owners_unsorted[0].address)
        self.assertTrue(signature_0_removed)
        self.assertEqual(set([]), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 0)

        # test unsign with invalid address
        safe_tx.sign(owners_unsorted[0].key)
        signers = [owner_addresses[0]]
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 1)

        signature_removed = safe_tx.unsign(Account.create().address)
        self.assertFalse(signature_removed)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 1)

    def test_hash_safe_multisig_tx(self):
        # -------- Old version of the contract --------------------------
        expected_hash = HexBytes(
            "0xc9d69a2350aede7978fdee58e702647e4bbdc82168577aa4a43b66ad815c6d1a"
        )
        tx_hash = SafeTx(
            self.ethereum_client,
            "0x692a70d2e424a56d2c6c27aa97d1a86395877b3a",
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5000000,
            HexBytes("0x00"),
            0,
            50000,
            100,
            10000,
            "0x" + "0" * 40,
            "0x" + "0" * 40,
            safe_nonce=67,
            safe_version="0.1.0",
        ).safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        expected_hash = HexBytes(
            "0x8ca8db91d72b379193f6e229eb2dff0d0621b6ef452d90638ee3206e9b7349b3"
        )
        tx_hash = SafeTx(
            self.ethereum_client,
            "0x692a70d2e424a56d2c6c27aa97d1a86395877b3a",
            "0x" + "0" * 40,
            80000000,
            HexBytes("0x562944"),
            2,
            54522,
            773,
            22000000,
            "0x" + "0" * 40,
            "0x" + "0" * 40,
            safe_nonce=257000,
            safe_version="0.1.0",
        ).safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        # -------- New version of the contract --------------------------
        expected_hash = HexBytes(
            "0x7c60341f3e1b4483575f38e84e97d6b332a2dd55b9290f39e6e26eef29a04fe7"
        )
        tx_hash = SafeTx(
            self.ethereum_client,
            "0x692a70D2e424a56D2C6C27aA97D1a86395877b3A",
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5000000,
            HexBytes("0x00"),
            0,
            50000,
            100,
            10000,
            "0x" + "0" * 40,
            "0x" + "0" * 40,
            safe_nonce=67,
            safe_version="1.0.0",
        ).safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        expected_hash = HexBytes(
            "0xf585279fd867c94738096f4eab964e9e202014d2f0d5155d751099ad85cbe504"
        )
        tx_hash = SafeTx(
            self.ethereum_client,
            "0x692a70D2e424a56D2C6C27aA97D1a86395877b3A",
            "0x" + "0" * 40,
            80000000,
            HexBytes("0x562944"),
            2,
            54522,
            773,
            22000000,
            "0x" + "0" * 40,
            "0x" + "0" * 40,
            safe_nonce=257000,
            safe_version="1.0.0",
        ).safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        safe = self.deploy_test_safe_v1_1_1()
        # Expected hash must be the same calculated by `getTransactionHash` of the contract
        expected_hash = safe.contract.functions.getTransactionHash(
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5212459,
            HexBytes(0x00),
            1,
            123456,
            122,
            12345,
            "0x" + "2" * 40,
            "0x" + "2" * 40,
            10789,
        ).call()
        safe_tx_hash = SafeTx(
            self.ethereum_client,
            safe.address,
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5212459,
            HexBytes(0x00),
            1,
            123456,
            122,
            12345,
            "0x" + "2" * 40,
            "0x" + "2" * 40,
            safe_nonce=10789,
            safe_version="1.0.0",
        ).safe_tx_hash
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)

        # Safe v1.3.0
        safe = self.deploy_test_safe()
        # Expected hash must be the same calculated by `getTransactionHash` of the contract
        expected_hash = safe.contract.functions.getTransactionHash(
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5212459,
            HexBytes(0x00),
            1,
            123456,
            122,
            12345,
            "0x" + "2" * 40,
            "0x" + "2" * 40,
            10789,
        ).call()
        safe_tx_hash = SafeTx(
            self.ethereum_client,
            safe.address,
            "0x5AC255889882aaB35A2aa939679E3F3d4Cea221E",
            5212459,
            HexBytes(0x00),
            1,
            123456,
            122,
            12345,
            "0x" + "2" * 40,
            "0x" + "2" * 40,
            safe_nonce=10789,
            safe_version="1.3.0",
        ).safe_tx_hash
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)

    def test_hash_safe_multisig_tx_1_3(self):
        safe_address = "0x2B0b9cBDA3D7b0F760187c52A8FFB18C48E5d96A"
        expected_hash = (
            "0x01d42a801ac44d3ebd43592be980dea0756d7f7b27d718d3016a42ea7ce85587"
        )
        to = "0x79613FD49472C3C7a32188e45ff00e7bdC8a897d"
        data = HexBytes("0x1212")
        value = 2
        operation = 0
        safe_tx_gas = 0
        base_gas = 0
        gas_price = 0
        gas_token = "0xFA995c7a0d32A4e7497508a5C380369BE8dB49Db"
        refund_receiver = "0x6b92f5E5360bfCa8F5d3FcE65D87382967847983"
        safe_nonce = 17
        chain_id = 4  # Rinkeby
        safe_tx_hash = SafeTx(
            self.ethereum_client,
            safe_address,
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            safe_nonce=safe_nonce,
            safe_version="1.3.0",
            chain_id=chain_id,
        ).safe_tx_hash
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)

        # Test with ganache master copy contract v1.3.0
        chain_id = self.ethereum_client.w3.eth.chain_id
        safe_tx_hash = SafeTx(
            self.ethereum_client,
            self.safe_contract.address,
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            safe_nonce=safe_nonce,
            safe_version="1.3.0",
            chain_id=chain_id,
        ).safe_tx_hash
        expected_hash = self.safe_contract.functions.getTransactionHash(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            safe_nonce,
        ).call()
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)



from hexbytes import HexBytes

# Entrypoint v0.6.0 and v0.7.0 events ----------------------------------------------------

# AccountDeployed (index_topic_1 bytes32 userOpHash, index_topic_2 address sender, address factory, address paymaster)
ACCOUNT_DEPLOYED_TOPIC = HexBytes(
    "0xd51a9c61267aa6196961883ecf5ff2da6619c37dac0fa92122513fb32c032d2d"
)

# Deposited (index_topic_1 address account, uint256 totalDeposit)
DEPOSIT_EVENT_TOPIC = HexBytes(
    "0x2da466a7b24304f47e87fa2e1e5a81b9831ce54fec19055ce277ca2f39ba42c4"
)

# Safe (L2 not required) >= 1.4.1 events -------------------------------------------------

# ExecutionFromModuleSuccess (index_topic_1 address module)
EXECUTION_FROM_MODULE_SUCCESS_TOPIC = HexBytes(
    "0x6895c13664aa4f67288b25d7a21d7aaa34916e355fb9b6fae0a139a9085becb8"
)

# ExecutionFromModuleFailure (index_topic_1 address module)
EXECUTION_FROM_MODULE_FAILURE_TOPIC = HexBytes(
    "0xacd2c8702804128fdb0db2bb49f6d127dd0181c13fd45dbfe16de0930e2bd375"
)



import logging
from unittest import mock

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes

from gnosis.eth import EthereumClient

from ..multi_send import MultiSend, MultiSendOperation, MultiSendTx
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestMultiSend(SafeTestCaseMixin, TestCase):
    def test_multisend_init(self):
        # Should try to detect the contract address
        with self.assertRaisesMessage(
            ValueError, "Cannot find a MultiSend contract for chainId=1337"
        ):
            MultiSend(self.ethereum_client)

        with mock.patch.object(EthereumClient, "is_contract", return_value=True):
            multisend = MultiSend(self.ethereum_client, call_only=False)
            self.assertEqual(multisend.address, multisend.MULTISEND_ADDRESSES[0])

            multisend = MultiSend(self.ethereum_client)
            self.assertEqual(
                multisend.address, multisend.MULTISEND_CALL_ONLY_ADDRESSES[0]
            )

        random_address = Account.create().address
        multisend = MultiSend(self.ethereum_client, address=random_address)
        self.assertEqual(multisend.address, random_address)

        # Check with no ethereum_client
        multisend = MultiSend(call_only=False)
        self.assertEqual(multisend.address, multisend.MULTISEND_ADDRESSES[0])

        multisend = MultiSend()
        self.assertEqual(multisend.address, multisend.MULTISEND_CALL_ONLY_ADDRESSES[0])

    def test_multi_send_tx_from_bytes(self):
        operation = MultiSendOperation.DELEGATE_CALL
        address = Account.create().address
        value = 876
        data = HexBytes("0x123456789a")
        multi_send_tx = MultiSendTx(operation, address, value, data)
        new_multi_send_tx = MultiSendTx.from_bytes(multi_send_tx.encoded_data)

        self.assertEqual(new_multi_send_tx, multi_send_tx)
        self.assertEqual(new_multi_send_tx.operation, operation)
        self.assertEqual(new_multi_send_tx.to, address)
        self.assertEqual(new_multi_send_tx.value, value)
        self.assertEqual(new_multi_send_tx.data, data)
        self.assertTrue(str(new_multi_send_tx))  # Test __str__

    def test_multi_send_from_bytes(self):
        encoded_multisend_txs = b""
        values = [876 * i for i in range(3)]
        datas = [HexBytes("0x123456789a") * i for i in range(3)]
        for value, data in zip(values, datas):  # Craft the same transaction three times
            operation = MultiSendOperation.DELEGATE_CALL
            address = Account.create().address
            multi_send_tx = MultiSendTx(operation, address, value, data)
            new_multi_send_tx = MultiSendTx.from_bytes(multi_send_tx.encoded_data)

            self.assertEqual(new_multi_send_tx, multi_send_tx)
            self.assertEqual(new_multi_send_tx.operation, operation)
            self.assertEqual(new_multi_send_tx.to, address)
            self.assertEqual(new_multi_send_tx.value, value)
            self.assertEqual(new_multi_send_tx.data, data)
            encoded_multisend_txs += multi_send_tx.encoded_data

        multisend_txs = MultiSend.from_bytes(encoded_multisend_txs)
        self.assertEqual(len(multisend_txs), 3)
        for multi_send_tx, value, data in zip(multisend_txs, values, datas):
            self.assertEqual(multi_send_tx.value, value)
            self.assertEqual(multi_send_tx.data, data)

    def test_multisend_parse_real_transaction(self):
        # Change Safe contract master copy and set fallback manager multisend transaction
        safe_contract_address = "0x5B9ea52Aaa931D4EEf74C8aEaf0Fe759434FeD74"
        data = HexBytes(
            "0x8d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000"
            "00000000000000000000000000000000000000000000f2005b9ea52aaa931d4eef74c8aeaf0fe759434fed740000"
            "00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
            "000000000000000000000000000000247de7edef00000000000000000000000034cfac646f301356faa8b21e9422"
            "7e3583fe3f5f005b9ea52aaa931d4eef74c8aeaf0fe759434fed7400000000000000000000000000000000000000"
            "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000024f0"
            "8a0323000000000000000000000000d5d82b6addc9027b22dca772aa68d5d74cdbdf440000000000000000000000"
            "000000"
        )
        change_master_copy_data = HexBytes(
            "0x7de7edef00000000000000000000000034cfac646f301356faa8b21e94227e3583fe3f5f"
        )
        change_fallback_manager_data = HexBytes(
            "0xf08a0323000000000000000000000000d5d82b6addc9027b22dca772aa68d5d74cd"
            "bdf44"
        )
        multisend_txs = MultiSend.from_transaction_data(data)
        self.assertEqual(len(multisend_txs), 2)
        for multisend_tx, expected_data in zip(
            multisend_txs, (change_master_copy_data, change_fallback_manager_data)
        ):
            self.assertEqual(multisend_tx.to, safe_contract_address)
            self.assertEqual(multisend_tx.data, expected_data)
            self.assertEqual(multisend_tx.value, 0)

    def test_multisend_parse_invalid_transaction(self):
        # Remove some data but keep the selector
        data = HexBytes(
            "0x8d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000"
            "00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
            "000000000000000000000000000000247de7edef00000000000000000000000034cfac646f301356faa8b21e9422"
            "7e3583fe3f5f005b9ea52aaa931d4eef74c8aeaf0fe759434fed7400000000000000000000000000000000000000"
            "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000024f0"
            "8a0323000000000000000000000000d5d82b6addc9027b22dca772aa68d5d74cdbdf440000000000000000000000"
            "000000"
        )
        multisend_txs = MultiSend.from_transaction_data(data)
        self.assertEqual(multisend_txs, [])

        # Keep the data but change the selector
        data = HexBytes(
            "0x1d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000"
            "00000000000000000000000000000000000000000000f2005b9ea52aaa931d4eef74c8aeaf0fe759434fed740000"
            "00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
            "000000000000000000000000000000247de7edef00000000000000000000000034cfac646f301356faa8b21e9422"
            "7e3583fe3f5f005b9ea52aaa931d4eef74c8aeaf0fe759434fed7400000000000000000000000000000000000000"
            "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000024f0"
            "8a0323000000000000000000000000d5d82b6addc9027b22dca772aa68d5d74cdbdf440000000000000000000000"
            "000000"
        )
        multisend_txs = MultiSend.from_transaction_data(data)
        self.assertEqual(multisend_txs, [])

    def test_multisend_parse_old_transaction(self):
        data = HexBytes(
            "0x8d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000024000000000000000000000000000000000000000000000000000000000000000000000000000000000000000001db1a2063504d15d36e9506a192ab7bbb17dbb2a000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000800000000000000000000000000000000000000000000000000000000000000024e71bdf41000000000000000000000000980f3fb56999bf811f32b85e5274960fbf9f5c4f0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000001db1a2063504d15d36e9506a192ab7bbb17dbb2a0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000008000000000000000000000000000000000000000000000000000000000000000a4beaeb388000000000000000000000000980f3fb56999bf811f32b85e5274960fbf9f5c4f00000000000000000000000089d24a6b4ccb1b6faa2625fe562bdd9a232603590000000000000000000000000000000000000000000000008ac7230489e8000000000000000000000000000000000000000000000000000000000000000005a000000000000000000000000000000000000000000000000000000000000005a000000000000000000000000000000000000000000000000000000000"
        )
        multisend_txs = MultiSend.from_transaction_data(data)
        self.assertEqual(len(multisend_txs), 2)
        self.assertEqual(multisend_txs[0].operation.value, 0)
        self.assertEqual(
            multisend_txs[0].to, "0x1DB1a2063504D15D36e9506A192ab7BBb17dbb2A"
        )
        self.assertEqual(multisend_txs[0].value, 0)
        self.assertEqual(
            multisend_txs[0].data,
            HexBytes(
                "0xe71bdf41000000000000000000000000980f3fb56999bf811f32b85e5274960fbf9f5c4f"
            ),
        )
        self.assertEqual(multisend_txs[0].data_length, 36)

        self.assertEqual(multisend_txs[1].operation.value, 0)
        self.assertEqual(
            multisend_txs[1].to, "0x1DB1a2063504D15D36e9506A192ab7BBb17dbb2A"
        )
        self.assertEqual(multisend_txs[1].value, 0)
        self.assertEqual(
            multisend_txs[1].data,
            HexBytes(
                "0xbeaeb388000000000000000000000000980f3fb56999bf811f32b85e5274960fbf9f5c4f00000000000000000000000089d24a6b4ccb1b6faa2625fe562bdd9a232603590000000000000000000000000000000000000000000000008ac7230489e8000000000000000000000000000000000000000000000000000000000000000005a000000000000000000000000000000000000000000000000000000000000005a0"
            ),
        )
        self.assertEqual(multisend_txs[1].data_length, 164)

    def test_multisend_build_tx_data(self):
        value = 4
        multisend_txs = [
            MultiSendTx(MultiSendOperation.CALL, Account.create().address, value, b"")
            for _ in range(2)
        ]

        # Test MultiSend with ethereum client/address and without
        for multi_send in (self.multi_send, MultiSend()):
            with self.subTest(multi_send=multi_send):
                safe_multisend_data = self.multi_send.build_tx_data(multisend_txs)
                self.assertGreater(len(safe_multisend_data), 0)
                self.assertEqual(
                    MultiSend.from_transaction_data(safe_multisend_data), multisend_txs
                )



from unittest import TestCase
from unittest.mock import patch

from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from gnosis.safe.api.transaction_service_api.transaction_service_messages import (
    get_delegate_message,
    get_remove_transaction_message,
    get_totp,
)


class TestTransactionServiceMessages(TestCase):
    def test_get_totp(self):
        mocked_time = 1615974000
        with patch("time.time", return_value=mocked_time):
            totp = get_totp()

        expected_totp = mocked_time // 3600
        self.assertEqual(totp, expected_totp)

    def test_get_delegate_message(self):
        chain_id = 1
        mocked_totp = 123
        delegate_address = ChecksumAddress("0x1234567890123456789012345678901234567890")

        expected_message = {
            "types": {
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                ],
                "Delegate": [
                    {"name": "delegateAddress", "type": "address"},
                    {"name": "totp", "type": "uint256"},
                ],
            },
            "primaryType": "Delegate",
            "domain": {
                "name": "Safe Transaction Service",
                "version": "1.0",
                "chainId": chain_id,
            },
            "message": {
                "delegateAddress": delegate_address,
                "totp": mocked_totp,
            },
        }

        with patch(
            "gnosis.safe.api.transaction_service_api.transaction_service_messages.get_totp",
            return_value=mocked_totp,
        ):
            message = get_delegate_message(delegate_address, chain_id)

        self.assertEqual(message, expected_message)

    def test_get_remove_transaction_message(self):
        chain_id = 1
        mocked_totp = 123
        delegate_address = ChecksumAddress("0x1234567890123456789012345678901234567890")
        safe_address = ChecksumAddress("0x1234567890123456789012345678901234567890")

        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )

        expected_message = {
            "types": {
                "EIP712Domain": [
                    {"name": "name", "type": "string"},
                    {"name": "version", "type": "string"},
                    {"name": "chainId", "type": "uint256"},
                    {"name": "verifyingContract", "type": "address"},
                ],
                "DeleteRequest": [
                    {"name": "safeTxHash", "type": "bytes32"},
                    {"name": "totp", "type": "uint256"},
                ],
            },
            "primaryType": "DeleteRequest",
            "domain": {
                "name": "Safe Transaction Service",
                "version": "1.0",
                "chainId": chain_id,
                "verifyingContract": safe_address,
            },
            "message": {
                "safeTxHash": safe_tx_hash,
                "totp": mocked_totp,
            },
        }

        with patch(
            "gnosis.safe.api.transaction_service_api.transaction_service_messages.get_totp",
            return_value=mocked_totp,
        ):
            message = get_remove_transaction_message(
                delegate_address, safe_tx_hash, chain_id
            )

        self.assertEqual(message, expected_message)



import logging

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes

from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS
from gnosis.eth.contracts import get_safe_contract, get_sign_message_lib_contract
from gnosis.eth.utils import fast_keccak_text, get_empty_tx_params

from ..enums import SafeOperationEnum
from ..exceptions import (
    CannotEstimateGas,
    CannotRetrieveSafeInfoException,
    CouldNotPayGasWithToken,
    InvalidInternalTx,
)
from ..safe import Safe, SafeV100, SafeV111, SafeV130, SafeV141
from ..signatures import signature_to_bytes, signatures_to_bytes
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestSafe(SafeTestCaseMixin, TestCase):
    @property
    def safe_contract(self):
        """
        :return: Last Safe Contract available
        """
        return self.safe_contract_V1_4_1

    def deploy_test_safe(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return: Deployed Safe Contract with the last version available
        """
        return super().deploy_test_safe_v1_4_1(*args, **kwargs)

    def test_create(self):
        owners = [self.ethereum_test_account.address]
        threshold = 1
        master_copy_address = self.safe_contract.address

        ethereum_tx_sent = Safe.create(
            self.ethereum_client,
            self.ethereum_test_account,
            master_copy_address,
            owners,
            threshold,
            proxy_factory_address=self.proxy_factory_contract.address,
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(safe.retrieve_master_copy_address(), master_copy_address)
        self.assertEqual(safe.retrieve_owners(), owners)
        self.assertEqual(safe.retrieve_threshold(), threshold)

    def test_domain_separator(self):
        safe = self.deploy_test_safe()
        self.assertTrue(safe.domain_separator)

        not_a_safe = Account.create().address
        self.assertIsNone(Safe(not_a_safe, self.ethereum_client).domain_separator)

    def test_inherit_safe(self):
        address = Account.create().address
        safe = SafeV141(address, self.ethereum_client)
        self.assertEqual(safe.address, address)
        self.assertEqual(safe.get_version(), "1.4.1")
        self.assertEqual(safe.chain_id, 1337)

        class CustomSafeV141(SafeV141):
            def get_version(self):
                return "1.4.1-custom"

        inherited_safe = CustomSafeV141(address, self.ethereum_client)
        self.assertEqual(inherited_safe.address, address)
        self.assertEqual(inherited_safe.get_version(), "1.4.1-custom")
        self.assertEqual(inherited_safe.chain_id, 1337)

    def test_check_funds_for_tx_gas(self):
        safe = self.deploy_test_safe()
        safe_tx_gas = 2
        base_gas = 4
        gas_price = 10
        self.assertFalse(
            safe.check_funds_for_tx_gas(safe_tx_gas, base_gas, gas_price, NULL_ADDRESS)
        )
        self.send_ether(safe.address, (safe_tx_gas + base_gas) * gas_price)
        self.assertTrue(
            safe.check_funds_for_tx_gas(safe_tx_gas, base_gas, gas_price, NULL_ADDRESS)
        )

    def test_estimate_safe_creation_2(self):
        number_owners = 4
        gas_price = self.gas_price
        payment_token = NULL_ADDRESS
        safe_creation_estimate = Safe.estimate_safe_creation_2(
            self.ethereum_client,
            self.safe_contract.address,
            self.proxy_factory_contract.address,
            number_owners,
            gas_price,
            payment_token,
        )
        self.assertGreater(safe_creation_estimate.gas_price, 0)
        self.assertGreater(safe_creation_estimate.gas, 0)
        self.assertGreater(safe_creation_estimate.payment, 0)
        self.assertEqual(safe_creation_estimate.payment_token, payment_token)

        salt_nonce = 167
        owners = [Account.create().address for _ in range(number_owners)]
        threshold = 1
        safe_creation_2 = Safe.build_safe_create2_tx(
            self.ethereum_client,
            self.safe_contract.address,
            self.proxy_factory_contract.address,
            salt_nonce,
            owners,
            threshold,
            gas_price,
            payment_token,
            fallback_handler=Account.create().address,
        )
        self.assertAlmostEqual(
            safe_creation_2.gas, safe_creation_estimate.gas, delta=1000
        )

    def test_retrieve_master_copy_address(self):
        master_copy_address = self.safe_contract.address
        ethereum_tx_send = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account, master_copy_address
        )
        safe = Safe(ethereum_tx_send.contract_address, self.ethereum_client)
        self.assertEqual(safe.retrieve_master_copy_address(), master_copy_address)

    def test_send_multisig_tx(self):
        # Create Safe
        w3 = self.w3
        funder_account = self.ethereum_test_account
        funder = funder_account.address
        accounts = [Account.create() for _ in range(2)]
        # Signatures must be sorted!
        accounts.sort(key=lambda a: a.address.lower())
        owners = [account.address for account in accounts]
        keys = [account.key for account in accounts]
        threshold = len(accounts)

        safe = self.deploy_test_safe(threshold=threshold, owners=owners)
        my_safe_address = safe.address

        # The balance we will send to the safe
        safe_balance = w3.to_wei(0.02, "ether")

        # Send something to the owner[0], who will be sending the tx
        owner0_balance = safe_balance
        self.send_tx({"to": owners[0], "value": owner0_balance}, funder_account)

        my_safe_contract = get_safe_contract(w3, my_safe_address)
        safe = Safe(my_safe_address, self.ethereum_client)

        to = funder
        value = safe_balance // 2
        data = HexBytes("")
        operation = 0
        safe_tx_gas = 100000
        base_gas = 300000
        gas_price = 1
        gas_token = NULL_ADDRESS
        refund_receiver = NULL_ADDRESS
        nonce = None
        safe_multisig_tx = safe.build_multisig_tx(
            to=to,
            value=value,
            data=data,
            operation=operation,
            safe_tx_gas=safe_tx_gas,
            base_gas=base_gas,
            gas_price=gas_price,
            gas_token=gas_token,
            refund_receiver=refund_receiver,
            safe_nonce=nonce,
        )
        safe_multisig_tx_hash = safe_multisig_tx.safe_tx_hash

        nonce = safe.retrieve_nonce()
        self.assertEqual(
            safe.build_multisig_tx(
                to=to,
                value=value,
                data=data,
                operation=operation,
                safe_tx_gas=safe_tx_gas,
                base_gas=base_gas,
                gas_price=gas_price,
                gas_token=gas_token,
                refund_receiver=refund_receiver,
                safe_nonce=nonce,
            ).safe_tx_hash,
            safe_multisig_tx_hash,
        )
        # Just to make sure we are not miscalculating tx_hash
        contract_multisig_tx_hash = my_safe_contract.functions.getTransactionHash(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            nonce,
        ).call()

        self.assertEqual(safe_multisig_tx_hash, contract_multisig_tx_hash)

        for private_key in keys:
            safe_multisig_tx.sign(private_key)

        signatures = safe_multisig_tx.signatures
        self.assertEqual(set(safe_multisig_tx.signers), set(owners))

        # Check owners are the same
        contract_owners = my_safe_contract.functions.getOwners().call()
        self.assertEqual(set(contract_owners), set(owners))
        self.assertEqual(w3.eth.get_balance(owners[0]), owner0_balance)

        # with self.assertRaises(CouldNotPayGasWithEther):
        # Ganache v7 does not raise CouldNotPayGasWithEther anymore
        with self.assertRaises(InvalidInternalTx):
            safe.send_multisig_tx(
                to,
                value,
                data,
                operation,
                safe_tx_gas,
                base_gas,
                gas_price,
                gas_token,
                refund_receiver,
                signatures,
                tx_sender_private_key=keys[0],
                tx_gas_price=self.gas_price,
            )

        # Send something to the safe
        self.send_tx({"to": my_safe_address, "value": safe_balance}, funder_account)

        ethereum_tx_sent = safe.send_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signatures,
            tx_sender_private_key=keys[0],
            tx_gas_price=self.gas_price,
        )

        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertTrue(tx_receipt["status"])
        owner0_new_balance = w3.eth.get_balance(owners[0])
        gas_used = tx_receipt["gasUsed"]
        gas_cost = gas_used * self.gas_price
        estimated_payment = (base_gas + gas_used) * gas_price
        real_payment = owner0_new_balance - (owner0_balance - gas_cost)
        # Estimated payment will be bigger, because it uses all the tx gas. Real payment only uses gas left
        # in the point of calculation of the payment, so it will be slightly lower
        self.assertTrue(estimated_payment > real_payment > 0)
        self.assertTrue(
            owner0_new_balance
            > owner0_balance - ethereum_tx_sent.tx["gas"] * self.gas_price
        )
        self.assertEqual(safe.retrieve_nonce(), 1)

    def test_send_multisig_tx_gas_token(self):
        # Create safe with one owner, fund the safe and the owner with `safe_balance`
        receiver = Account.create().address
        threshold = 1
        funder_account = self.ethereum_test_account
        funder = funder_account.address
        safe_balance_ether = 0.02
        safe_balance = self.w3.to_wei(safe_balance_ether, "ether")
        owner_account = self.create_and_fund_account(initial_ether=safe_balance_ether)
        owner = owner_account.address

        safe = self.deploy_test_safe(
            threshold=threshold, owners=[owner], initial_funding_wei=safe_balance
        )
        my_safe_address = safe.address

        # Give erc20 tokens to the funder
        amount_token = int(1e18)
        erc20_contract = self.deploy_example_erc20(amount_token, funder)
        self.assertEqual(
            self.ethereum_client.erc20.get_balance(funder, erc20_contract.address),
            amount_token,
        )

        signature_packed = signature_to_bytes(1, int(owner, 16), 0)

        to = receiver
        value = safe_balance
        data = HexBytes("")
        operation = 0
        safe_tx_gas = 100000
        base_gas = 300000
        gas_price = 2
        gas_token = erc20_contract.address
        refund_receiver = NULL_ADDRESS

        with self.assertRaises(CouldNotPayGasWithToken):
            safe.send_multisig_tx(
                to,
                value,
                data,
                operation,
                safe_tx_gas,
                base_gas,
                gas_price,
                gas_token,
                refund_receiver,
                signature_packed,
                tx_sender_private_key=owner_account.key,
                tx_gas_price=self.gas_price,
            )

        # Give erc20 tokens to the safe
        self.ethereum_client.erc20.send_tokens(
            my_safe_address, amount_token, erc20_contract.address, funder_account.key
        )

        safe.send_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signature_packed,
            tx_sender_private_key=owner_account.key,
            tx_gas_price=self.gas_price,
        )

        safe_token_balance = self.ethereum_client.erc20.get_balance(
            my_safe_address, erc20_contract.address
        )

        # Token was used for tx gas costs. Sender must have some tokens now and safe should have less
        self.assertLess(safe_token_balance, amount_token)
        owner_token_balance = self.ethereum_client.erc20.get_balance(
            owner, erc20_contract.address
        )
        self.assertGreater(owner_token_balance, 0)

        # All ether on safe was transferred to receiver
        receiver_balance = self.w3.eth.get_balance(receiver)
        self.assertEqual(receiver_balance, safe_balance)

    def test_estimate_tx_base_gas(self):
        safe = self.deploy_test_safe()
        to = Account().create().address
        value = int("abc", 16)
        data = HexBytes("0xabcdef")
        operation = 1
        gas_token = NULL_ADDRESS
        estimate_tx_gas = int("ccdd", 16)
        base_gas = safe.estimate_tx_base_gas(
            to, value, data, operation, gas_token, estimate_tx_gas
        )
        self.assertGreater(base_gas, 0)

        data = HexBytes(
            "0xabcdefbb"
        )  # A byte that was 00 now is bb, so -4 + GAS_CALL_DATA_BYTE
        data_gas2 = safe.estimate_tx_base_gas(
            to, value, data, operation, gas_token, estimate_tx_gas
        )
        self.assertEqual(data_gas2, base_gas + GAS_CALL_DATA_BYTE - 4)

    def test_estimate_tx_gas(self):
        to = Account().create().address
        value = 123
        data = HexBytes("0xabcdef")
        operation = 1
        safe = self.deploy_test_safe(initial_funding_wei=value + 23000)

        safe_tx_gas = safe.estimate_tx_gas(to, value, data, operation)
        self.assertGreater(safe_tx_gas, 0)
        operation = 0
        safe_tx_gas = safe.estimate_tx_gas(to, value, data, operation)
        self.assertGreater(safe_tx_gas, 0)

    def test_estimate_tx_gas_nested_transaction(self):
        """
        Test estimation with a contract that uses a lot of gas
        """

        # Uncomment this to regenerate the contract. I hardcoded it to prevent the app testing depending on solc
        # Remember to install `py-solc` and have `solidity` in your system
        """
        from solc import compile_standard
        compiled_sol = compile_standard({
            "language": "Solidity",
            "sources": {
                "Nester.sol": {
                    "content": '''
                        contract Nester {
                            uint256[] public data;
                            constructor() public payable {}
                            function nested(uint256 level, uint256 count) external {
                                if (level == 0) {
                                    for (uint256 i = 0; i < count; i++) {
                                        data.push(i);
                                    }
                                    return;
                                }
                                this.nested(level - 1, count);
                            }
                            function useGas(uint256 count) public {
                                this.nested(6, count);
                                this.nested(8, count);
                            }
                        }
                    '''
                }
            },
            "settings":
                {
                    "outputSelection": {

                        "*": {
                            "*": [
                                "metadata", "evm.bytecode"
                                , "evm.bytecode.sourceMap"
                            ]
                        }
                    }
                }
        })
        contract_data = compiled_sol['contracts']['Nester.sol']['Nester']
        bytecode = contract_data['evm']['bytecode']['object']
        abi = json.loads(contract_data['metadata'])['output']['abi']
        """

        bytecode = "60806040526102fe806100136000396000f3fe608060405234801561001057600080fd5b50600436106100415760003560e01c8063022952b81461004657806350d1f08214610074578063f0ba8440146100ac575b600080fd5b6100726004803603602081101561005c57600080fd5b81019080803590602001909291905050506100ee565b005b6100aa6004803603604081101561008a57600080fd5b8101908080359060200190929190803590602001909291905050506101d9565b005b6100d8600480360360208110156100c257600080fd5b81019080803590602001909291905050506102a7565b6040518082815260200191505060405180910390f35b3073ffffffffffffffffffffffffffffffffffffffff166350d1f0826006836040518363ffffffff1660e01b81526004018083815260200182815260200192505050600060405180830381600087803b15801561014a57600080fd5b505af115801561015e573d6000803e3d6000fd5b505050503073ffffffffffffffffffffffffffffffffffffffff166350d1f0826008836040518363ffffffff1660e01b81526004018083815260200182815260200192505050600060405180830381600087803b1580156101be57600080fd5b505af11580156101d2573d6000803e3d6000fd5b5050505050565b600082141561022c5760008090505b8181101561022657600081908060018154018082558091505060019003906000526020600020016000909190919091505580806001019150506101e8565b506102a3565b3073ffffffffffffffffffffffffffffffffffffffff166350d1f08260018403836040518363ffffffff1660e01b81526004018083815260200182815260200192505050600060405180830381600087803b15801561028a57600080fd5b505af115801561029e573d6000803e3d6000fd5b505050505b5050565b600081815481106102b457fe5b90600052602060002001600091509050548156fea264697066735822122091b08fac39dd94b262ed9adf68b679a88140319e98d18d4c918bd5a1d93527fc64736f6c63430006040033"
        abi = [
            {"inputs": [], "stateMutability": "payable", "type": "constructor"},
            {
                "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                "name": "data",
                "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                "stateMutability": "view",
                "type": "function",
            },
            {
                "inputs": [
                    {"internalType": "uint256", "name": "level", "type": "uint256"},
                    {"internalType": "uint256", "name": "count", "type": "uint256"},
                ],
                "name": "nested",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
            {
                "inputs": [
                    {"internalType": "uint256", "name": "count", "type": "uint256"}
                ],
                "name": "useGas",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]

        nester_contract = self.w3.eth.contract(abi=abi, bytecode=bytecode)
        tx_hash = nester_contract.constructor().transact(
            {"from": self.w3.eth.accounts[0]}
        )
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        nester = self.w3.eth.contract(address=tx_receipt["contractAddress"], abi=abi)

        safe = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address],
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        nester_tx = nester.functions.useGas(80).build_transaction(
            {"gasPrice": 1, "from": safe.address, "gas": 1}
        )
        nester_data = nester_tx["data"]

        safe_tx_gas_no_call = safe.estimate_tx_gas_with_safe(
            nester.address, 0, nester_data, SafeOperationEnum.CALL.value
        )
        safe_tx_gas = safe.estimate_tx_gas_by_trying(
            nester.address, 0, nester_data, SafeOperationEnum.CALL.value
        )
        self.assertGreater(safe_tx_gas, safe_tx_gas_no_call)

        base_gas = safe.estimate_tx_base_gas(
            nester.address,
            0,
            nester_data,
            SafeOperationEnum.CALL.value,
            NULL_ADDRESS,
            safe_tx_gas,
        )

        refund_receiver = Account.create().address
        safe_tx = safe.build_multisig_tx(
            nester.address,
            0,
            nester_data,
            safe_tx_gas=safe_tx_gas,
            base_gas=base_gas,
            gas_price=1,
            refund_receiver=refund_receiver,
        )
        safe_tx.sign(self.ethereum_test_account.key)

        self.assertTrue(
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address)
        )

        safe_tx_hash, safe_w3_tx = safe_tx.execute(self.ethereum_test_account.key)
        self.w3.eth.wait_for_transaction_receipt(safe_tx_hash)

        # Tx was successfully executed if refund_receiver gets ether
        self.assertGreater(self.ethereum_client.get_balance(refund_receiver), 0)

    def test_estimate_tx_gas_with_web3(self):
        safe = self.deploy_test_safe(
            owners=[self.ethereum_test_account.address],
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        to = Account.create().address
        value = self.w3.to_wei(0.01, "ether")
        data = b""
        gas_estimated_web3 = safe.estimate_tx_gas_with_web3(to, value, data)
        gas_estimated_safe = safe.estimate_tx_gas_with_safe(to, value, data, 0)
        self.assertGreater(
            gas_estimated_safe, gas_estimated_web3
        )  # Web3 estimation should use less gas
        self.assertGreater(gas_estimated_web3, 0)
        self.assertGreater(gas_estimated_safe, 0)

        with self.assertRaises(CannotEstimateGas):
            deployed_erc20 = self.deploy_example_erc20(100, Account.create().address)
            transfer_data = deployed_erc20.functions.transfer(
                to, 200
            ).build_transaction(get_empty_tx_params())["data"]
            safe.estimate_tx_gas_with_web3(deployed_erc20.address, value, transfer_data)

    def test_retrieve_code(self):
        self.assertEqual(
            Safe(NULL_ADDRESS, self.ethereum_client).retrieve_code(), HexBytes("0x")
        )
        self.assertIsNotNone(self.deploy_test_safe().retrieve_code())

    def test_retrieve_fallback_handler(self):
        random_fallback_handler = Account.create().address
        safe = self.deploy_test_safe(fallback_handler=random_fallback_handler)
        self.assertEqual(safe.retrieve_fallback_handler(), random_fallback_handler)

    def test_retrieve_guard(self):
        owner_account = Account.create()
        safe = self.deploy_test_safe(owners=[owner_account.address])
        self.assertEqual(safe.retrieve_guard(), NULL_ADDRESS)

        guard_address = self.deploy_example_guard()
        set_guard_data = HexBytes(
            safe.contract.functions.setGuard(guard_address).build_transaction(
                get_empty_tx_params()
            )["data"]
        )
        set_guard_tx = safe.build_multisig_tx(safe.address, 0, set_guard_data)
        set_guard_tx.sign(owner_account.key)
        set_guard_tx.execute(self.ethereum_test_account.key)
        self.assertEqual(safe.retrieve_guard(), guard_address)

    def test_retrieve_info(self):
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        safe = self.deploy_test_safe(owners=owners, threshold=threshold)

        self.assertEqual(
            safe.retrieve_master_copy_address(), self.safe_contract.address
        )
        self.assertEqual(safe.retrieve_nonce(), 0)
        self.assertCountEqual(safe.retrieve_owners(), owners)
        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertEqual(safe.retrieve_modules(), [])

        # Versions must be semantic, like 0.1.0, so we count 3 points
        self.assertTrue(safe.retrieve_version().count("."), 3)

        for owner in owners:
            self.assertTrue(safe.retrieve_is_owner(owner))

    def test_retrieve_all_info(self):
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        safe = self.deploy_test_safe(owners=owners, threshold=threshold)

        safe_info = safe.retrieve_all_info()
        self.assertEqual(safe_info.master_copy, self.safe_contract.address)
        self.assertEqual(safe_info.nonce, 0)
        self.assertCountEqual(safe_info.owners, owners)
        self.assertEqual(safe_info.threshold, threshold)
        self.assertEqual(safe_info.modules, [])

        invalid_address = Account.create().address
        invalid_safe = Safe(invalid_address, self.ethereum_client)
        with self.assertRaisesMessage(CannotRetrieveSafeInfoException, invalid_address):
            invalid_safe.retrieve_all_info()

    def test_safe_instance(self):
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        safe_v1_0_0 = self.deploy_test_safe_v1_0_0(owners=owners, threshold=threshold)
        self.assertTrue(isinstance(safe_v1_0_0, SafeV100))
        safe_v1_1_1 = self.deploy_test_safe_v1_1_1(owners=owners, threshold=threshold)
        self.assertTrue(isinstance(safe_v1_1_1, SafeV111))
        safe_v1_3_0 = self.deploy_test_safe_v1_3_0(owners=owners, threshold=threshold)
        self.assertTrue(isinstance(safe_v1_3_0, SafeV130))
        safe_v1_4_1 = self.deploy_test_safe_v1_4_1(owners=owners, threshold=threshold)
        self.assertTrue(isinstance(safe_v1_4_1, SafeV141))

    def test_retrieve_modules_new(self):
        safe = self.deploy_test_safe(owners=[self.ethereum_test_account.address])
        safe_contract = safe.contract
        module_address = Account.create().address
        self.assertEqual(safe.retrieve_modules(), [])

        tx = safe_contract.functions.enableModule(module_address).build_transaction(
            {"from": self.ethereum_test_account.address, "gas": 0, "gasPrice": 0}
        )
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx["data"])
        safe_tx.sign(self.ethereum_test_account.key)
        safe_tx.execute(
            tx_sender_private_key=self.ethereum_test_account.key,
            tx_gas_price=self.gas_price,
        )
        self.assertEqual(safe.retrieve_modules(), [module_address])

        more_modules = [Account.create().address for _ in range(2)]
        for more_module in more_modules:
            # Test pagination
            tx = safe_contract.functions.enableModule(more_module).build_transaction(
                {"from": self.ethereum_test_account.address, "gas": 0, "gasPrice": 0}
            )
            safe_tx = safe.build_multisig_tx(safe.address, 0, tx["data"])
            safe_tx.sign(self.ethereum_test_account.key)
            safe_tx.execute(
                tx_sender_private_key=self.ethereum_test_account.key,
                tx_gas_price=self.gas_price,
            )
        self.assertCountEqual(
            safe.retrieve_modules(pagination=1), [module_address] + more_modules
        )

    def test_retrieve_modules_unitialized_safe(self):
        """
        Unitialized Safes from V1.4.1 will revert
        """

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            self.ethereum_test_account,
            self.safe_contract.address,
            initializer=b"",
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        with self.assertRaisesMessage(ValueError, "revert"):
            self.assertEqual(safe.retrieve_modules(), [])
        self.assertEqual(safe.retrieve_all_info().modules, [])

    def test_retrieve_is_hash_approved(self):
        safe = self.deploy_test_safe(owners=[self.ethereum_test_account.address])
        safe_contract = safe.contract
        fake_tx_hash = fast_keccak_text("Knopfler")
        another_tx_hash = fast_keccak_text("Marc")
        tx = safe_contract.functions.approveHash(fake_tx_hash).build_transaction(
            {"from": self.ethereum_test_account.address}
        )

        self.ethereum_client.send_unsigned_transaction(
            tx, private_key=self.ethereum_test_account.key
        )
        self.assertTrue(
            safe.retrieve_is_hash_approved(
                self.ethereum_test_account.address, fake_tx_hash
            )
        )
        self.assertFalse(
            safe.retrieve_is_hash_approved(
                self.ethereum_test_account.address, another_tx_hash
            )
        )

    def test_retrieve_is_message_signed_v1_1_1(self):
        safe = self.deploy_test_safe_v1_1_1(owners=[self.ethereum_test_account.address])
        safe_contract = safe.contract
        message = b"12345"
        message_hash = safe.get_message_hash(message)
        self.assertFalse(safe.retrieve_is_message_signed(message_hash))
        sign_message_data = HexBytes(
            safe_contract.functions.signMessage(message).build_transaction(
                get_empty_tx_params()
            )["data"]
        )
        safe_tx = safe.build_multisig_tx(safe.address, 0, sign_message_data)
        safe_tx.sign(self.ethereum_test_account.key)
        safe_tx.execute(tx_sender_private_key=self.ethereum_test_account.key)
        self.assertTrue(safe.retrieve_is_message_signed(message_hash))

    def test_retrieve_is_message_signed(self):
        safe = self.deploy_test_safe(owners=[self.ethereum_test_account.address])
        sign_message_lib_address = self.deploy_sign_message_lib()
        sign_message_contract = get_sign_message_lib_contract(self.w3, safe.address)
        message = b"12345"
        message_hash = safe.get_message_hash(message)
        self.assertFalse(safe.retrieve_is_message_signed(message_hash))
        sign_message_data = sign_message_contract.functions.signMessage(
            message
        ).build_transaction(get_empty_tx_params())["data"]
        safe_tx = safe.build_multisig_tx(
            sign_message_lib_address,
            0,
            sign_message_data,
            SafeOperationEnum.DELEGATE_CALL.value,
        )
        safe_tx.sign(self.ethereum_test_account.key)
        safe_tx.execute(tx_sender_private_key=self.ethereum_test_account.key)
        self.assertTrue(safe.retrieve_is_message_signed(message_hash))

    def test_retrieve_is_owner(self):
        safe = self.deploy_test_safe(owners=[self.ethereum_test_account.address])
        self.assertTrue(safe.retrieve_is_owner(self.ethereum_test_account.address))
        self.assertFalse(safe.retrieve_is_owner(Account.create().address))

    def test_token_balance(self):
        funder_account = self.ethereum_test_account
        funder = funder_account.address
        amount = 200
        deployed_erc20 = self.deploy_example_erc20(amount, funder)

        safe = self.deploy_test_safe(threshold=2, number_owners=3)
        my_safe_address = safe.address

        balance = self.ethereum_client.erc20.get_balance(
            my_safe_address, deployed_erc20.address
        )
        self.assertEqual(balance, 0)

        transfer_tx = deployed_erc20.functions.transfer(
            my_safe_address, amount
        ).build_transaction({"from": funder})
        self.send_tx(transfer_tx, funder_account)

        balance = self.ethereum_client.erc20.get_balance(
            my_safe_address, deployed_erc20.address
        )
        self.assertEqual(balance, amount)

    # TODO Test approve tx from another contract
    def test_send_previously_approved_tx(self):
        number_owners = 4
        accounts = [
            self.create_and_fund_account(initial_ether=0.01)
            for _ in range(number_owners)
        ]
        accounts.sort(key=lambda x: x.address.lower())
        owners = [account.address for account in accounts]

        safe = self.deploy_test_safe(
            threshold=2,
            owners=owners,
            initial_funding_wei=self.w3.to_wei(0.01, "ether"),
        )
        safe_address = safe.address
        safe = Safe(safe_address, self.ethereum_client)
        safe_instance = get_safe_contract(self.w3, safe_address)

        to = Account.create().address
        value = self.w3.to_wei(0.001, "ether")
        data = b""
        operation = 0
        safe_tx_gas = 500000
        data_gas = 500000
        gas_price = 1
        gas_token = NULL_ADDRESS
        refund_receiver = NULL_ADDRESS
        nonce = safe.retrieve_nonce()

        self.assertEqual(nonce, 0)

        safe_tx_hash = safe.build_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            data_gas,
            gas_price,
            gas_token,
            refund_receiver,
            safe_nonce=nonce,
        ).safe_tx_hash

        safe_tx_contract_hash = safe_instance.functions.getTransactionHash(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            data_gas,
            gas_price,
            gas_token,
            refund_receiver,
            nonce,
        ).call()

        self.assertEqual(safe_tx_hash, safe_tx_contract_hash)

        approve_hash_fn = safe_instance.functions.approveHash(safe_tx_hash)
        for account in accounts[:2]:
            self.send_tx(
                approve_hash_fn.build_transaction({"from": account.address}), account
            )

        for owner in (owners[0], owners[1]):
            is_approved = safe.retrieve_is_hash_approved(owner, safe_tx_hash)
            self.assertTrue(is_approved)

        # Prepare signatures. v must be 1 for previously signed and r the owner
        signatures = (1, int(owners[0], 16), 0), (1, int(owners[1], 16), 0)
        signature_bytes = signatures_to_bytes(signatures)

        safe.send_multisig_tx(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            data_gas,
            gas_price,
            gas_token,
            refund_receiver,
            signature_bytes,
            self.ethereum_test_account.key,
        )

        balance = self.w3.eth.get_balance(to)
        self.assertEqual(value, balance)
